import numpy as np
import os

import astropy.units as u
from astropy import constants as const
from astropy.table import QTable

from glue.core.message import (SubsetCreateMessage,
                               SubsetDeleteMessage,
                               SubsetUpdateMessage)
from glue_jupyter.common.toolbar_vuetify import read_icon
from traitlets import Any, Bool, Float, Int, List, Unicode, Dict, observe

from jdaviz.core.custom_traitlets import FloatHandleEmpty
from jdaviz.core.events import (AddDataMessage,
                                RemoveDataMessage,
                                AddLineListMessage,
                                LineIdentifyMessage,
                                SnackbarMessage,
                                RedshiftMessage,
                                SpectralMarksChangedMessage)
from jdaviz.core.linelists import load_preset_linelist, get_linelist_metadata
from jdaviz.core.marks import SpectralLine
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import PluginTemplateMixin
from jdaviz.core.tools import ICON_DIR
from jdaviz.core.unit_conversion_utils import create_equivalent_spectral_axis_units_list

__all__ = ['LineListTool']


@tray_registry(
    'g-line-list', label="Line Lists",
    viewer_requirements=['spectrum']
)
class LineListTool(PluginTemplateMixin):
    dialog = Bool(False).tag(sync=True)
    template_file = __file__, "line_lists.vue"

    rs_enabled = Bool(False).tag(sync=True)  # disabled until lines are plotted
    rs_slider = Float(0).tag(sync=True)  # in units of delta-redshift
    rs_slider_range_auto = Bool(True).tag(sync=True)
    rs_slider_half_range = Float(0.1).tag(sync=True)
    rs_slider_step_auto = Bool(True).tag(sync=True)
    rs_slider_step = Float(0.01).tag(sync=True)
    rs_slider_ndigits = Int(1).tag(sync=True)
    rs_slider_throttle = Int(100).tag(sync=True)
    rs_redshift = FloatHandleEmpty(0).tag(sync=True)
    rs_rv = FloatHandleEmpty(0).tag(sync=True)
    rs_rv_step = Float(1).tag(sync=True)

    dc_items = List([]).tag(sync=True)
    available_lists = List([]).tag(sync=True)
    loaded_lists = List([]).tag(sync=True)
    list_contents = Dict({}).tag(sync=True)
    custom_name = Unicode().tag(sync=True)
    custom_rest = Unicode().tag(sync=True)
    custom_unit_choices = List([]).tag(sync=True)
    custom_unit = Unicode().tag(sync=True)

    lines_filter = Any().tag(sync=True)  # string or None
    filter_range = Bool(False).tag(sync=True)
    spectrum_viewer_min = Float(0.01).tag(sync=True)
    spectrum_viewer_max = Float(0.01).tag(sync=True)

    identify_label = Unicode().tag(sync=True)
    identify_line_icon = Unicode(read_icon(os.path.join(ICON_DIR, 'line_select.svg'), 'svg+xml')).tag(sync=True)  # noqa
    filter_range_icon = Unicode(read_icon(os.path.join(ICON_DIR, 'spectral_range.svg'), 'svg+xml')).tag(sync=True)  # noqa

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._viewer = self.app.get_viewer(self._default_spectrum_viewer_reference_name)
        self._spectrum1d = None
        if self._viewer:
            self.available_lists = self._viewer.available_linelists()
        self.list_to_load = None
        self.loaded_lists = ["Custom"]
        self.list_contents = {"Custom": {"lines": [],
                                         "color": "#FF0000FF",
                                         "medium": "Unknown (Custom)"}}
        self.line_mark_dict = {}
        self._units = {}
        self._bounds = {}
        self._global_redshift = 0
        self._rs_disable_observe = False
        self._rs_pause_tables = False
        # track which line was recently changed to avoid recursive updates due to imprecise
        # roundtripping
        self._rs_line_obs_change = (None, None)

        # Watch for messages from Specviz helper redshift functions
        self.hub.subscribe(self, RedshiftMessage,
                           handler=self._parse_redshift_msg)

        self.hub.subscribe(self, AddDataMessage,
                           handler=self._on_viewer_data_changed)

        self.hub.subscribe(self, RemoveDataMessage,
                           handler=self._on_viewer_data_changed)

        self.hub.subscribe(self, SubsetCreateMessage,
                           handler=lambda x: self._on_viewer_data_changed())

        self.hub.subscribe(self, SubsetDeleteMessage,
                           handler=lambda x: self._on_viewer_data_changed())

        self.hub.subscribe(self, SubsetUpdateMessage,
                           handler=lambda x: self._on_viewer_data_changed())

        self.hub.subscribe(self, AddLineListMessage,
                           handler=self._list_from_notebook)

        self.hub.subscribe(self, LineIdentifyMessage,
                           handler=self._process_identify_change)

        self.hub.subscribe(self, SpectralMarksChangedMessage,
                           handler=lambda x: self.update_line_mark_dict())

        # if set to auto (default), update the slider range when zooming on the spectrum viewer
        if self._viewer:
            self._viewer.state.add_callback("x_min",
                                            lambda x_min: self._on_spectrum_viewer_limits_changed())
            self._viewer.state.add_callback("x_max",
                                            lambda x_max: self._on_spectrum_viewer_limits_changed())

        self._disable_if_no_data()

        # description displayed under plugin title in tray
        self._plugin_description = 'Plot spectral lines from preset or custom line lists.'

    @property
    def _default_spectrum_viewer_reference_name(self):
        return getattr(
            self.app._jdaviz_helper, '_default_spectrum_viewer_reference_name', 'spectrum-viewer'
        )

    @property
    def _default_flux_viewer_reference_name(self):
        return getattr(
            self.app._jdaviz_helper, '_default_flux_viewer_reference_name', 'flux-viewer'
        )

    def _disable_if_no_data(self):
        if len(self.app.data_collection) == 0:
            self.disabled_msg = 'Line Lists unavailable when no data is loaded'
        else:
            self.disabled_msg = ''

    def _on_viewer_data_changed(self, msg=None):
        """
        Callback method for when data is added or removed from a viewer, or
        when a subset is created, deleted, or updated. This method receives
        a glue message containing viewer information in the case of the former
        set of events, and updates the units in which to display the lines.

        Notes
        -----
        We do not attempt to parse any data at this point, at it can cause
        visible lag in the application.

        Parameters
        ----------
        msg : `glue.core.Message`
            The glue message passed to this callback method.
        """
        self._disable_if_no_data()

        self._viewer_id = self.app._viewer_item_by_reference(
            self._default_spectrum_viewer_reference_name).get('id')

        # Subsets are global and are not linked to specific viewer instances,
        # so it's not required that we match any specific ids for that case.
        # However, if the msg is not none, check to make sure that it's the
        # viewer we care about and that the message contains the data label.
        if msg is None or msg.viewer_id != self._viewer_id or msg.data is None:
            return

        viewer = self.app.get_viewer(self._default_spectrum_viewer_reference_name)
        viewer_data_labels = [layer.layer.label for layer in viewer.layers]
        if msg.data.label not in viewer_data_labels:
            return

        label = msg.data.label
        try:
            viewer_data = self.app._jdaviz_helper.get_data(data_label=label)
        except TypeError:
            warn_message = SnackbarMessage("Line list plugin could not retrieve data from viewer",
                                           sender=self, color="error")
            self.hub.broadcast(warn_message)
            return

        # If no data is currently plotted, don't attempt to update
        if viewer_data is None:
            return

        if viewer_data.spectral_axis.unit == u.pix:
            # disable the plugin until we can address this properly (either using the wavelength
            # solution to support pixels in line-lists, or properly displaying the extracted
            # 1d spectrum in wavelength-space)
            self.disabled_msg = 'Line Lists unavailable when x-axis is in pixels'
        else:
            self.disabled_msg = ''

        self._units["x"] = str(viewer_data.spectral_axis.unit)
        self._units["y"] = str(viewer_data.flux.unit)

        self._bounds["min"] = viewer_data.spectral_axis[0]
        self._bounds["max"] = viewer_data.spectral_axis[-1]

        # set redshift slider to redshift stored in Spectrum1D object
        if viewer_data.meta.get('plugin', None) is not None:
            self.rs_redshift = (viewer_data.redshift.value
                                if hasattr(viewer_data.redshift, 'value')
                                else viewer_data.redshift)
        self._on_spectrum_viewer_limits_changed()  # will also trigger _auto_slider_step

        # set the choices (and default) for the units for new custom lines
        self.custom_unit_choices = create_equivalent_spectral_axis_units_list(
            viewer_data.spectral_axis.unit)
        self.custom_unit = str(viewer_data.spectral_axis.unit)

    def _parse_redshift_msg(self, msg):
        '''
        Handle incoming redshift messages from the app hub. Generally these
        will be created by Specviz helper methods.
        '''
        if msg.sender == self:
            return

        param = msg.param

        if param == "rs_slider_range":
            if msg.value == 'auto':
                # observer will handle setting rs_slider_range
                self.rs_slider_range_auto = True
            else:
                self.rs_slider_range_auto = False
                self.rs_slider_half_range = float(msg.value)/2
        elif param == "rs_slider_step":
            if msg.value == 'auto':
                # observer will handle setting rs_slider_step
                self.rs_slider_step_auto = True
            else:
                self.rs_slider_step_auto = False
                slider_step = float(msg.value)
                if slider_step > self.rs_slider_half_range:
                    raise ValueError("step must be smaller than range/2")
                self.rs_slider_step = slider_step
                self.rs_rv_step = self._redshift_to_velocity(slider_step)
        elif param == "redshift":
            # NOTE: this should trigger the observe to update rs_rv, line positions, and
            # update self._global_redshift
            self.rs_redshift = float(msg.value)
        elif param == 'rv':
            # NOTE: this should trigger the observe to update rs_redshift, line positions, and
            # update self._global_redshift
            self.rs_rv = float(msg.value)
        else:
            raise NotImplementedError(f"RedshiftMessage with param {param} not implemented.")

    def _velocity_to_redshift(self, velocity):
        """
        Convert a velocity to a relativistic redshift.  Assumes km/s (float)
        as input and returns float.
        """
        # NOTE: if supporting non-km/s units in the future, try to leave
        # the default case to avoid quantity math as below for efficiency
        beta = velocity * 1000 / const.c.value
        return np.sqrt((1 + beta) / (1 - beta)) - 1

    def _redshift_to_velocity(self, redshift):
        """
        Convert a relativistic redshift to a velocity.  Returns
        in km/s (float)
        """
        zponesq = (1 + redshift) ** 2
        # NOTE: if supporting non-km/s units in the future, try to leave
        # the default case to avoid quantity math as below for efficiency
        return const.c.value * (zponesq - 1) / (zponesq + 1) / 1000  # km/s

    def _update_line_positions(self):
        # update all lines, self._global_redshift, and emit message back to Specviz helper
        z = u.Quantity(self.rs_redshift)

        for mark in self.app.get_viewer(self._default_spectrum_viewer_reference_name).figure.marks:
            # update ALL to this redshift, if adding support for per-line redshift
            # this logic will need to change to not affect ALL lines
            if not isinstance(mark, SpectralLine):
                continue

            mark.redshift = z

    @observe('rs_slider')
    def _on_slider_updated(self, event):
        if self._rs_disable_observe:
            return

        self._rs_pause_tables = True

        # NOTE: _on_rs_redshift_updated will handle updating rs_rv
        # NOTE: the input has a custom @input method in line_lists.vue to cast
        # to float so that we can assume its a float here to minimize lag
        # when interacting with the slider.

        self.rs_redshift = np.round(self.rs_redshift + event['new'] - event['old'],
                                    self.rs_slider_ndigits)

    def _rest_to_obs(self, rest, redshift=None):
        if redshift is None:
            redshift = float(self.rs_redshift)

        return rest * (1+redshift)

    @observe('rs_redshift')
    def _on_rs_redshift_updated(self, event):
        if self._rs_disable_observe:
            return

        if not isinstance(event['new'], float):
            # then blank or None or '.'
            return

        value = event['new']
        # update _global_redshift so new lines, etc, will adopt this latest value
        self._global_redshift = value
        self._rs_disable_observe = True
        self.rs_rv = self._redshift_to_velocity(value)
        self._rs_disable_observe = False
        self._update_line_positions()

        if not self._rs_pause_tables:
            # TODO: try to avoid essentially repeating the loop from above, careful to minimize
            # updates to vue, maybe pause traitlets?
            self._update_line_list_obs()

            # Send the redshift back to the Specviz helper (and also trigger
            # self._update_global_redshift)
            msg = RedshiftMessage("redshift", value, sender=self)
            self.app.hub.broadcast(msg)

    def _update_line_list_obs(self, *args):
        for list_name, line_list in self.list_contents.items():
            for i, line in enumerate(line_list['lines']):
                if self._rs_line_obs_change[0] == list_name and self._rs_line_obs_change[1] == i:  # noqa
                    # this trigger is coming from a manual change to the observed
                    # wavelength and would result in a small change to the value before the
                    # user can finish typing.  So we'll just keep the old value until the
                    # widget is blurred (loses focus)
                    line_list['lines'][i]['obs'] = self._rs_line_obs_change[2]
                else:
                    line_list['lines'][i]['obs'] = self._rest_to_obs(float(line['rest']))

            self.list_contents[list_name] = line_list

        self.send_state('list_contents')

    def vue_change_line_obs(self, kwargs):
        # NOTE: we can only pass one argument from vue (it seems), so we'll pass as
        # a dictionary (kwargs) instead of positional or keyword arguments (**kwargs)
        line_obs = kwargs.get('obs_new')
        if isinstance(line_obs, str) and not len(line_obs):
            # empty string, we don't want to revert yet because then
            # the user can never delete the entry and type something new
            # so we'll just leave empty
            return
        list_name = kwargs.get('list_name')
        line_ind = kwargs.get('line_ind')
        line = self.list_contents[list_name]['lines'][line_ind]
        line_rest = float(line['rest'])
        if line_obs is None:
            # then coming from the blur, we'll keep the latest update from the @change
            line_obs = float(line['obs'])

        # we don't want this call to recursively update THIS obs wavelength, but DO want it to
        # update the RV and all other obs wavelengths.  Once tabbing or losing focus, vue will
        # send another event with avoid_feedback=False so that the wavelength updates to
        # exactly match the redshift (so that can be considered the ground truth value consistently)
        if kwargs.get('avoid_feedback', False):
            self._rs_line_obs_change = (list_name, line_ind, line_obs)
        # ensure tables will update when rs_redshift change is observed
        self._rs_pause_tables = kwargs.get('avoid_feedback', False)
        self.rs_redshift = (line_obs - line_rest) / line_rest
        self._rs_line_obs_change = (None, None)

    def vue_unpause_tables(self, event=None):
        # after losing focus, update any elements that were paused during changes
        self._rs_pause_tables = False
        self._rs_disable_observe = False
        self._on_rs_redshift_updated({'new': self.rs_redshift})

    @observe('rs_rv')
    def _on_rs_rv_updated(self, event):
        if self._rs_disable_observe:
            return

        if not isinstance(event['new'], float):
            # then blank or None or '.'
            return

        value = event['new']
        redshift = self._velocity_to_redshift(value)
        # prevent update the redshift from propagating back to an update in the rv
        self._rs_disable_observe = True
        # we'll wait until the blur event (which will call vue_unpause_tables)
        # to update the value in the MOS table and observed wavelengths
        self._rs_pause_tables = True
        self.rs_redshift = redshift
        # but we do want to update the plotted lines
        self._update_line_positions()

        self._rs_disable_observe = False

    def vue_slider_reset(self, event):
        self._rs_disable_observe = True
        self.rs_slider = 0.0
        self._rs_disable_observe = False
        self._rs_pause_tables = False
        # the redshift value in the MOS table and observed wavelengths weren't
        # updating during slide, so update them now
        self.vue_unpause_tables()

    def _on_spectrum_viewer_limits_changed(self, event=None):
        sv = self.app.get_viewer(self._default_spectrum_viewer_reference_name)
        if sv.state.x_min is None or sv.state.x_max is None:
            return
        self.spectrum_viewer_min = float(sv.state.x_min)
        self.spectrum_viewer_max = float(sv.state.x_max)

        # Also update the slider range
        self._auto_slider_range()

    def _auto_slider_range(self, event=None):
        """
        Automatically adjusts the Redshift slider range to the values of the
        spectrum_viewer_min and spectrum_viewer_max traitlets
        """
        if not self.rs_slider_range_auto:
            return
        # if set to auto, default the range based on the limits of the spectrum plot
        x_min, x_max = self.spectrum_viewer_min, self.spectrum_viewer_max
        x_mid = abs(x_max + x_min) / 2.
        # we'll *estimate* the redshift range to shift the range of the viewer
        # (for a line with a rest wavelength in the center of the viewer),
        # by taking abs, this will work for wavelength or frequency units.
        if x_mid == 0:
            # logic below fails, leave the half-range as-is
            # instead of raising a divide by zero error
            return
        half_range = abs(x_max - x_min) / x_mid
        ndec = -np.log10(half_range)
        if ndec > 0 and not np.isinf(ndec):
            # round to at least 2 digits, or the first significant digit
            ndec = np.max([2, int(np.ceil(ndec))])
        else:
            ndec = 1
        half_range = np.round(half_range, ndec)

        # this will trigger self._auto_slider_step to set self.rs_slider_step and
        # self.rs_rv_step, if applicable
        self.rs_slider_half_range = half_range

    @observe('rs_slider_range_auto')
    def _on_rs_slider_range_auto_updated(self, event):
        if event['new']:
            self._on_spectrum_viewer_limits_changed()

    @observe('rs_slider_half_range')
    def _auto_slider_step(self, event=None):
        if not self.rs_slider_step_auto:
            return
        # if set to auto, default to 1000 steps in the range
        self.rs_slider_step = self.rs_slider_half_range * 2 / 1000
        self.rs_rv_step = abs(self._redshift_to_velocity(self._global_redshift+self.rs_slider_step) - self.rs_rv) # noqa

    @observe('rs_slider_step')
    def _on_rs_slider_step_updated(self, event):
        # When using the slider, we'll "round" redshift to the digits in the
        # slider step to avoid extra digits due to rounding errors
        ndec = -np.log10(event['new'])
        if ndec > 0 and not np.isinf(ndec):
            # round to at least 2 digits, or one past the first significant digit
            # note: the UI will not show trailing zeros, we just want to avoid
            # and 1 at floating point precision if not significant
            ndec = np.max([2, np.ceil(ndec)+1])
        else:
            ndec = 1
        self.rs_slider_ndigits = int(ndec)

    @observe('rs_slider_step_auto')
    def _on_rs_slider_step_auto_updated(self, event):
        if event['new']:
            self._auto_slider_step()

    def _update_global_redshift(self, msg):
        '''Handle updates to the Specviz redshift slider, to apply to lines'''
        if msg.param == "redshift":
            self._global_redshift = msg.value

    def _list_from_notebook(self, msg):
        """
        Callback method for when a spectral line list is added to the specviz
        instance from the notebook.

        Parameters
        ----------
        msg : `glue.core.Message`
            The glue message passed to this callback method. Includes the line
            data added in msg.table.
        """
        loaded_lists = self.loaded_lists
        list_contents = self.list_contents
        tmp_names_rest = []
        for row in msg.table:
            if row["listname"] not in loaded_lists:
                loaded_lists.append(row["listname"])

            if row["listname"] not in list_contents:
                list_contents[row["listname"]] = {"lines": [], "color": "#FF0000FF"}

            temp_dict = {"linename": row["linename"],
                         "rest": row["rest"].value,
                         "obs": self._rest_to_obs(row["rest"].value),
                         "unit": str(row["rest"].unit),
                         "colors": row["colors"] if "colors" in row else "#FF0000FF",
                         "show": row["show"],
                         "name_rest": row["name_rest"]}
            list_contents[row["listname"]]["lines"].append(temp_dict)
            tmp_names_rest.append(row["name_rest"])

        self.send_state('loaded_lists')
        self.send_state('list_contents')

        self._viewer.plot_spectral_lines(tmp_names_rest, global_redshift=self._global_redshift)
        self.update_line_mark_dict()

        msg_text = ("Spectral lines loaded from notebook. Lines can be hidden"
                    "/shown in the Line Lists plugin")
        lines_loaded_message = SnackbarMessage(msg_text, sender=self,
                                               color="success", timeout=15000)
        self.hub.broadcast(lines_loaded_message)

    def update_line_mark_dict(self):
        self.line_mark_dict = {}
        for m in self._viewer.figure.marks:
            if isinstance(m, SpectralLine):
                self.line_mark_dict[m.table_index] = m

        n_lines_shown = len(self.line_mark_dict)

        # redshift controls are enabled if any lines are currently plotted
        self.rs_enabled = n_lines_shown > 0

        if n_lines_shown > 0:
            # with a lot of lines, a quick slider move will lag.  Let's scale the
            # timeout based on the number of lines, roughtly between 50-500 ms
            throttle = n_lines_shown * 5
            if throttle < 50:
                throttle = 50
            if throttle > 500:
                throttle = 500
            self.rs_slider_throttle = throttle

    def vue_list_selected(self, event):
        """
        Handle list selection from presets dropdown selector
        """
        self.list_to_load = event

    def vue_load_list(self, event):
        """
        Load one of the preset line lists, storing it's info in a
        vuetify-friendly manner in addition to loading the astropy table into
        the viewer's spectral_lines attribute.
        """
        # Don't need to reload an already loaded list
        if self.list_to_load in self.loaded_lists:
            return
        temp_table = load_preset_linelist(self.list_to_load)

        # Also store basic list contents in a form that vuetify can handle
        # Adds line style parameters that can be changed on the front end
        temp_table["colors"] = "#FF0000FF"

        # Load the table into the main astropy table and get it back, to make
        # sure all values match between the main table and local plugin
        temp_table = self._viewer.load_line_list(temp_table, return_table=True,
                                                 show=False)

        metadata = get_linelist_metadata()
        list_medium = metadata[self.list_to_load].get('medium', 'Unknown').capitalize()

        line_list_dict = {"lines": [], "color": "#FF000080", "medium": list_medium}
        # extra_fields = [x for x in temp_table.colnames if x not in
        #                ("linename", "rest", "name_rest")]

        for row in temp_table:
            temp_dict = {"linename": row["linename"],
                         "rest": row["rest"].value,
                         "obs": self._rest_to_obs(row["rest"].value),
                         "unit": str(row["rest"].unit),
                         "colors": row["colors"],
                         "show": False,
                         "name_rest": str(row["name_rest"])}
            # for field in extra_fields:
            #     temp_dict[field] = row[field]
            line_list_dict["lines"].append(temp_dict)

        list_contents = self.list_contents
        list_contents[self.list_to_load] = line_list_dict
        self.list_contents = {}
        self.list_contents = list_contents

        loaded_lists = self.loaded_lists + [self.list_to_load]
        self.loaded_lists = []
        self.loaded_lists = loaded_lists

        self._viewer.plot_spectral_lines()
        self.update_line_mark_dict()

        msg_text = ("Spectral lines loaded from preset. Lines can be shown/hidden"
                    f" in the {self.list_to_load} dropdown in the Line Lists plugin")
        lines_loaded_message = SnackbarMessage(msg_text, sender=self,
                                               color="success", timeout=15000)
        self.hub.broadcast(lines_loaded_message)

    def vue_add_custom_line(self, event):
        """
        Add a line to the "Custom" line list from UI input
        """
        list_contents = self.list_contents
        temp_dict = {"linename": self.custom_name,
                     "rest": float(self.custom_rest),
                     "obs": self._rest_to_obs(float(self.custom_rest)),
                     "unit": self.custom_unit,
                     "colors": list_contents["Custom"]["color"],
                     "show": True
                     }

        # Add to viewer astropy table

        with u.set_enabled_equivalencies(u.spectral()):
            temp_table = QTable()
            temp_table["linename"] = [temp_dict["linename"]]
            temp_table["rest"] = [temp_dict["rest"]*u.Unit(temp_dict["unit"])]
            temp_table["colors"] = [temp_dict["colors"]]
            temp_table = self._viewer.load_line_list(temp_table, return_table=True)

            # Add line to Custom lines in local list
            temp_dict["name_rest"] = str(temp_table[0]["name_rest"])
            list_contents["Custom"]["lines"].append(temp_dict)
            self.list_contents = {}
            self.list_contents = list_contents

            self._viewer.plot_spectral_line(temp_dict["name_rest"])
            self.update_line_mark_dict()

        lines_loaded_message = SnackbarMessage("Custom spectral line loaded",
                                               sender=self, color="success")
        self.hub.broadcast(lines_loaded_message)

    def vue_show_all_in_list(self, listname):
        """
        Toggle all lines in list to be visible
        """
        lc = self.list_contents
        for line in lc[listname]["lines"]:
            line["show"] = True
            self._viewer.spectral_lines.loc[line["name_rest"]]["show"] = True

        self.list_contents = lc
        self.send_state('list_contents')

        self._viewer.plot_spectral_lines(global_redshift=self._global_redshift)
        self.update_line_mark_dict()

    def vue_hide_all_in_list(self, listname):
        """
        Toggle all lines in list to be hidden
        """
        name_rests = []
        for line in self.list_contents[listname]["lines"]:
            line["show"] = False
            name_rests.append(line["name_rest"])

        self.send_state('list_contents')

        self._viewer.erase_spectral_lines(name_rest=name_rests)
        self.update_line_mark_dict()

    def vue_plot_all_lines(self, event):
        """
        Plot all the currently loaded lines in the viewer
        """
        if self._viewer.spectral_lines is None:
            warn_message = SnackbarMessage("No spectral lines loaded to plot",
                                           sender=self, color="error")
            self.hub.broadcast(warn_message)
            return
        for listname in self.list_contents:
            for line in self.list_contents[listname]["lines"]:
                line["show"] = True
        self._viewer.spectral_lines["show"] = True

        self.send_state('list_contents')

        self._viewer.plot_spectral_lines(global_redshift=self._global_redshift)
        self.update_line_mark_dict()

    def vue_erase_all_lines(self, event):
        """
        Erase all lines from the viewer
        """
        if self._viewer.spectral_lines is None:
            warn_message = SnackbarMessage("No spectral lines to erase",
                                           sender=self, color="error")
            self.hub.broadcast(warn_message)
            return
        for listname in self.list_contents:
            for line in self.list_contents[listname]["lines"]:
                line["show"] = False

        self.send_state('list_contents')

        self._viewer.erase_spectral_lines()
        self.update_line_mark_dict()

    def vue_change_visible(self, data):
        """
        Plot or erase a single line as needed when "Visible" checkbox is changed
        """
        listname, line, line_ind = data
        name_rest = line["name_rest"]
        show = not line['show']

        list_contents = self.list_contents
        list_contents[listname]['lines'][line_ind]['show'] = show
        if not show:
            # then make sure to also disable the identify flag
            list_contents[listname]['lines'][line_ind]['identify'] = False
        self.list_contents = {}
        self.list_contents = list_contents

        if show:
            self._viewer.plot_spectral_line(name_rest, global_redshift=self._global_redshift)
        else:
            self._viewer.erase_spectral_lines(name_rest=name_rest)

        self.update_line_mark_dict()

    def _update_identify_to_line(self, name_rest, listname=None, identify=True):
        list_contents = self.list_contents
        for this_listname, this_list in list_contents.items():
            for i, line in enumerate(this_list['lines']):
                if ((this_listname == listname or listname is None) and
                        line['name_rest'] == name_rest):
                    list_contents[this_listname]['lines'][i]['identify'] = identify
                else:
                    list_contents[this_listname]['lines'][i]['identify'] = False

        self.list_contents = {}
        self.list_contents = list_contents
        self.identify_label = name_rest if identify else ""

    def _process_identify_change(self, msg):
        if msg.sender == self:
            return
        # event from some other plugin (LineAnalysis, for example) requesting a change
        # in the identified line
        self._update_identify_to_line(msg.name_rest)
        # then line mark themselves will also respond to the same event, so there is
        # no need to broadcast another

    def vue_set_identify(self, data=None):
        """
        Set the selected line as "identified"
        """
        if data is None:
            # then default to the currently identified (which will unidentify it)
            for listname, this_list in self.list_contents.items():
                for line_ind, line in enumerate(this_list['lines']):
                    if line['identify']:
                        return self.vue_set_identify((listname, line, line_ind))

        listname, line, line_ind = data
        identify = not line.get('identify', False)
        if identify and not line['show']:
            # first show the line
            self.vue_change_visible(data)

        self._update_identify_to_line(name_rest=line['name_rest'],
                                      listname=listname,
                                      identify=identify)

        # broadcast and event to update the marks
        msg = LineIdentifyMessage(name_rest=line['name_rest'] if identify else '',
                                  sender=self)
        self.hub.broadcast(msg)

    def vue_set_color(self, data):
        """
        Change the color either of all members of a line list, or of an
        individual line.
        """
        color = data['color']
        if "listname" in data:
            listname = data["listname"]

            self.list_contents[listname]["color"] = color

            for line in self.list_contents[listname]["lines"]:
                line["colors"] = color
                # Update the astropy table entry
                name_rest = line["name_rest"]
                self._viewer.spectral_lines.loc[name_rest]["colors"] = color
                # Update the color on the plot
                if name_rest in self.line_mark_dict:
                    self.line_mark_dict[name_rest].colors = [color]

            self.send_state('list_contents')

        elif "linename" in data:
            pass

    def vue_remove_list(self, listname):
        """
        Method to remove line list from available expansion panels when the x
        on the panel header is clicked. Also removes line marks from plot and
        updates the "show" value in the astropy table to False.
        """
        lc = self.list_contents[listname]
        name_rests = []
        for line in lc["lines"]:
            name_rests.append(self.vue_remove_line(line, erase=False))
        self._viewer.erase_spectral_lines(name_rest=name_rests)
        self.update_line_mark_dict()

        self.loaded_lists = [x for x in self.loaded_lists if x != listname]
        self.list_contents = {k: v for k, v in self.list_contents.items() if k != listname}
        row_inds = [i for i, ln in
                    enumerate(self._viewer.spectral_lines['listname'])
                    if ln != listname]
        if len(row_inds) == 0:
            self._viewer.spectral_lines = None
        else:
            self._viewer.spectral_lines = self._viewer.spectral_lines[row_inds]

    def vue_remove_line(self, line, erase=True):
        """
        Method to remove a line from the plot when the line is deselected in
        the expansion panel content. Input must have "linename" and "rest"
        values for indexing on the astropy table.
        """
        name_rest = line["name_rest"]
        # Keep in our spectral line astropy table, but set it to not show on plot
        self._viewer.spectral_lines.loc[name_rest]["show"] = False

        # Remove the line from the plot marks
        if erase:
            try:
                self._viewer.erase_spectral_lines(name_rest=name_rest)
                del self.line_mark_dict[name_rest]
            except KeyError:
                raise KeyError("line marks: {}".format(self._viewer.figure.marks))
        else:
            return name_rest



import pytest
from regions import EllipsePixelRegion, PixCoord
from specutils import SpectralRegion

from jdaviz import Cubeviz
from jdaviz.app import Application
from jdaviz.core.config import get_configuration


def test_nested_helper(cubeviz_helper):
    '''Ensures the Cubeviz helper is always returned, even after the Specviz helper is called'''
    # Force Specviz helper to instantiate
    cubeviz_helper.specviz

    assert cubeviz_helper.app._jdaviz_helper == cubeviz_helper

    # The viewers also have a callthrough to the app's _jdaviz_helper attribute
    spec_viewer = cubeviz_helper.app.get_viewer('spectrum-viewer')
    assert spec_viewer.jdaviz_helper == cubeviz_helper


# Some API might be going through deprecation, so ignore the warning.
@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_plugin_user_apis(cubeviz_helper):
    for plugin_name, plugin_api in cubeviz_helper.plugins.items():
        plugin = plugin_api._obj
        for attr in plugin_api._expose:
            assert hasattr(plugin, attr), f"{attr} not in {plugin_name}"


def test_remote_server_disable_save_serverside():
    config = get_configuration('cubeviz')
    config['settings']['server_is_remote'] = True

    cubeviz_app = Application(config)
    cubeviz_helper = Cubeviz(cubeviz_app)
    exp = cubeviz_helper.plugins['Export']
    assert 'mp4' not in exp.viewer_format.choices

    mm = cubeviz_helper.plugins['Moment Maps']
    assert mm._obj.export_enabled is False


def test_get_data_spatial_and_spectral(cubeviz_helper, spectrum1d_cube_larger):
    cubeviz_helper.load_data(spectrum1d_cube_larger, data_label="test")
    unit = spectrum1d_cube_larger.spectral_axis.unit
    subset_plugin = cubeviz_helper.plugins['Subset Tools']
    subset_plugin.import_region([
        EllipsePixelRegion(center=PixCoord(x=4.5, y=4), width=9, height=8),  # Subset 1 (spatial)
        SpectralRegion(4.62440061e-07 * unit, 4.62520112e-07 * unit),  # Subset 2 (spectral)
    ], combination_mode="new")
    spatial_with_spec = cubeviz_helper.get_data(data_label="Spectrum (Subset 1, sum)",
                                                spectral_subset="Subset 2")
    assert spatial_with_spec.flux.ndim == 1
    assert list(spatial_with_spec.mask) == [True, True, False, False, True,
                                            True, True, True, True, True]
    assert max(list(spatial_with_spec.flux.value)) == 232.
    assert min(list(spatial_with_spec.flux.value)) == 16.



import os
import re
import warnings
from copy import deepcopy

import numpy as np
from astropy.utils import deprecated
from glue.core.link_helpers import LinkSame

from jdaviz.core.events import SnackbarMessage, NewViewerMessage
from jdaviz.core.helpers import ImageConfigHelper
from jdaviz.utils import (data_has_valid_wcs, get_wcs_only_layer_labels,
                          get_reference_image_data, _wcs_only_label)

__all__ = ['Imviz']


class Imviz(ImageConfigHelper):
    """Imviz Helper class."""
    _default_configuration = 'imviz'
    _default_viewer_reference_name = "image-viewer"

    def create_image_viewer(self, viewer_name=None):
        """Create a new image viewer.

        To display data in this new viewer programmatically,
        first get the new viewer ID from the small tab on the top
        left of viewer display. Then, use
        :meth:`~jdaviz.app.Application.add_data_to_viewer` from ``imviz.app``
        by passing in the new viewer ID and the desired data label,
        once per dataset you wish to display.

        Alternately, you can also display data interactively via the GUI.

        Parameters
        ----------
        viewer_name : str or `None`
            Viewer name/ID to use. If `None`, it is auto-generated.

        Returns
        -------
        viewer : `~jdaviz.configs.imviz.plugins.viewers.ImvizImageView`
            Image viewer instance.

        """
        from jdaviz.configs.imviz.plugins.viewers import ImvizImageView

        # Cannot assign data to real Data because it loads but it will
        # not update checkbox in Data menu.

        # add WCS-only layers from all viewers into the new viewer
        add_layers_to_viewer = get_wcs_only_layer_labels(self.app)

        return self.app._on_new_viewer(
            NewViewerMessage(ImvizImageView, data=None, sender=self.app),
            vid=viewer_name, name=viewer_name,
            add_layers_to_viewer=add_layers_to_viewer)

    def destroy_viewer(self, viewer_id):
        """Destroy a viewer associated with the given ID.

        Raises
        ------
        ValueError
            Default viewer cannot be destroyed.

        """
        if viewer_id not in self.app._viewer_store:  # Silent no-op
            return
        if viewer_id == f'{self.app.config}-0':
            raise ValueError(f"Default viewer '{viewer_id}' cannot be destroyed")
        self.app.vue_destroy_viewer_item(viewer_id)

    def load_data(self, data, data_label=None, show_in_viewer=True, **kwargs):
        """Load data into Imviz.

        Parameters
        ----------
        data : obj or str
            File name or object to be loaded. Supported formats include:

            * ``'filename.fits'`` (or any extension that ``astropy.io.fits``
              supports; first image extension found is loaded unless ``ext``
              keyword is also given)
            * ``'filename.fits[SCI]'`` (loads only first SCI extension)
            * ``'filename.fits[SCI,2]'`` (loads the second SCI extension)
            * ``'filename.jpg'`` (requires ``scikit-image``; grayscale only)
            * ``'filename.png'`` (requires ``scikit-image``; grayscale only)
            * JWST ASDF-in-FITS file (requires ``stdatamodels`` and ``gwcs``; ``data`` or given
              ``ext`` + GWCS)
            * Roman ASDF file or `roman_datamodels.datamodels.ImageModel`
              (requires ``roman-datamodels``)
            * `~astropy.io.fits.HDUList` object (first image extension found
              is loaded unless ``ext`` keyword is also given)
            * `~astropy.io.fits.ImageHDU` object
            * `~astropy.nddata.NDData` object (2D only but may have unit,
              mask, or uncertainty attached)
            * Numpy array (2D or 3D); if 3D, it will treat each slice at
              ``axis=0`` as a separate image (limit is 16 slices), however
              loading too many slices will cause performance issue,
              so consider using Cubeviz instead.

        data_label : str or `None`
            Data label to go with the given data. If not given, this is
            automatically determined from filename or randomly generated.
            The final label shown in Imviz may have additional information
            appended for clarity.

        show_in_viewer : str or bool
            If `True`, show the data in default viewer.  If a string, show in that viewer.

        kwargs : dict
            Extra keywords to be passed into app-level parser.
            The only one you might call directly here is ``ext`` (any FITS
            extension format supported by `astropy.io.fits`).

        Notes
        -----
        When loading image formats that support RGB color like JPG or PNG, the
        files are converted to greyscale. This is done following the algorithm
        of :func:`skimage.color.rgb2gray`, which involves weighting the channels as
        ``0.2125 R + 0.7154 G + 0.0721 B``. If you prefer a different weighting,
        you can use :func:`skimage.io.imread` to produce your own greyscale
        image as Numpy array and load the latter instead.

        """
        prev_data_labels = self.app.data_collection.labels

        if isinstance(data, str):
            filelist = data.split(',')

            if len(filelist) > 1 and data_label:
                raise ValueError('Do not manually overwrite data_label for '
                                 'a list of images')

            for data in filelist:
                kw = deepcopy(kwargs)
                filepath, ext, cur_data_label = split_filename_with_fits_ext(data)

                # This, if valid, will overwrite input.
                if ext is not None:
                    kw['ext'] = ext

                # This will only overwrite if not provided.
                if not data_label:
                    kw['data_label'] = None
                else:
                    kw['data_label'] = data_label
                self.app.load_data(filepath, parser_reference='imviz-data-parser', **kw)

        elif isinstance(data, np.ndarray) and data.ndim >= 3:
            if data.ndim > 3:
                data = data.squeeze()
                if data.ndim != 3:
                    raise ValueError(f'Imviz cannot load this array with ndim={data.ndim}')

            max_n_slice = 16  # Arbitrary limit for performance reasons
            for i in range(data.shape[0]):
                if i == max_n_slice:
                    warnings.warn(f'{max_n_slice} or more 3D slices found, stopping; '
                                  'please use Cubeviz')
                    break

                kw = deepcopy(kwargs)

                if data_label:
                    kw['data_label'] = data_label

                self.app.load_data(data[i, :, :], parser_reference='imviz-data-parser', **kw)

        else:
            if data_label:
                kwargs['data_label'] = data_label
            self.app.load_data(data, parser_reference='imviz-data-parser', **kwargs)

        # find the current label(s) - TODO: replace this by calling default label functionality
        # above instead of having to refind it
        applied_labels = []
        applied_visible = []
        layer_is_wcs_only = []
        layer_has_wcs = []
        for data in self.app.data_collection:
            label = data.label
            if label not in prev_data_labels:
                applied_labels.append(label)
                applied_visible.append(True)
                layer_is_wcs_only.append(data.meta.get(_wcs_only_label, False))
                layer_has_wcs.append(data_has_valid_wcs(data))

        if show_in_viewer is True:
            show_in_viewer = f"{self.app.config}-0"

        if show_in_viewer:
            linked_by_wcs = self.app._align_by == 'wcs'
            if linked_by_wcs:
                for applied_label, visible, is_wcs_only, has_wcs in zip(
                        applied_labels, applied_visible, layer_is_wcs_only, layer_has_wcs
                ):
                    if not is_wcs_only and linked_by_wcs and not has_wcs:
                        self.app.hub.broadcast(SnackbarMessage(
                            f"'{applied_label}' will be added to the data collection but not "
                            f"the viewer '{show_in_viewer}', since data are linked by WCS, but "
                            f"'{applied_label}' has no WCS.",
                            color="warning", timeout=8000, sender=self)
                        )

        if self._in_batch_load and show_in_viewer:
            for applied_label, is_wcs_only in zip(applied_labels, layer_is_wcs_only):
                if not is_wcs_only:
                    self._delayed_show_in_viewer_labels[applied_label] = show_in_viewer

        else:
            if 'Orientation' not in self.plugins.keys():
                # otherwise plugin will handle linking automatically with DataCollectionAddMessage
                self.link_data(align_by='wcs')

            # One input might load into multiple Data objects.
            # NOTE: If the batch_load context manager was used, it will
            # handle that logic instead.
            if show_in_viewer:
                for applied_label, visible, has_wcs in zip(
                        applied_labels, applied_visible, layer_has_wcs
                ):
                    if (has_wcs and linked_by_wcs) or not linked_by_wcs:
                        self.app.add_data_to_viewer(show_in_viewer, applied_label, visible=visible)

    def link_data(self, align_by='pixels', wcs_fallback_scheme=None, wcs_fast_approximation=True):
        """(Re)link loaded data in Imviz with the desired link type.
        All existing links will be replaced.

        Parameters
        ----------
        align_by : {'pixels', 'wcs'}
            Choose to link by pixels or WCS.

        wcs_fallback_scheme : {None, 'pixels'}
            If WCS linking failed, choose to fall back to linking by pixels or not at all.
            This is only used when ``align_by='wcs'``.
            Choosing `None` may result in some Imviz functionality not working properly.

        wcs_fast_approximation : bool
            Use an affine transform to represent the offset between images if possible
            (requires that the approximation is accurate to within 1 pixel with the
            full WCS transformations). If approximation fails, it will automatically
            fall back to full WCS transformation. This is only used when ``align_by='wcs'``.
            Affine approximation is much more performant at the cost of accuracy.

        """
        from jdaviz.configs.imviz.plugins.orientation.orientation import align_by_msg_to_trait
        plg = self.plugins["Orientation"]
        plg._obj.wcs_use_fallback = wcs_fallback_scheme == 'pixels'
        plg.wcs_fast_approximation = wcs_fast_approximation
        plg.align_by = align_by_msg_to_trait[align_by]

    @deprecated(since="4.0", alternative="get_alignment_method")
    def get_link_type(self, data_label_1, data_label_2):
        return self.get_alignment_method(data_label_1, data_label_2)

    def get_alignment_method(self, data_label_1, data_label_2):
        """Find the type of ``glue`` linking between the given
        data labels. A link is bi-directional. If there are
        more than 2 data in the collection, one of the given
        labels should be the reference data or look-up will fail.

        Parameters
        ----------
        data_label_1, data_label_2 : str
           Labels for the data linked together.

        Returns
        -------
        align_by : {'pixels', 'wcs', 'self'}
            One of the link types accepted by the Orientation plugin
            or ``'self'`` if the labels are identical.

        Raises
        ------
        ValueError
            Link look-up failed.

        """
        if data_label_1 == data_label_2:
            return "self"

        align_by = None
        for elink in self.app.data_collection.external_links:
            elink_labels = (elink.data1.label, elink.data2.label)
            if data_label_1 in elink_labels and data_label_2 in elink_labels:
                if isinstance(elink, LinkSame):  # Assumes WCS link never uses LinkSame
                    align_by = 'pixels'
                else:  # If not pixels, must be WCS
                    align_by = 'wcs'
                break  # Might have duplicate, just grab first match

        if align_by is None:
            avail_links = [f"({elink.data1.label}, {elink.data2.label})"
                           for elink in self.app.data_collection.external_links]
            raise ValueError(f'{data_label_1} and {data_label_2} combo not found '
                             f'in data collection external links: {avail_links}')

        return align_by

    @deprecated(since="4.2", alternative="plugins['Aperture Photometry'].export_table()")
    def get_aperture_photometry_results(self):
        """Return aperture photometry results, if any.
        Results are calculated using :ref:`aper-phot-simple` plugin.

        Returns
        -------
        results : `~astropy.table.QTable` or `None`
            Photometry results if available or `None` otherwise.

        """
        return self.plugins['Aperture Photometry'].export_table()

    @deprecated(since="4.2", alternative="plugins['Catalog Search'].export_table()")
    def get_catalog_source_results(self):
        """Return table of sources given by querying from a catalog, if any.
        Results are calculated using :ref:`imviz-catalogs` plugin.

        Returns
        -------
        results : `~astropy.table.QTable` or `None`
            Table of sources if available or `None` otherwise.

        """
        return getattr(self.app, '_catalog_source_table', None)

    def get_data(self, data_label=None, spatial_subset=None, cls=None):
        """
        Returns data with name equal to data_label of type cls with subsets applied from
        spatial_subset.

        Parameters
        ----------
        data_label : str, optional
            Provide a label to retrieve a specific data set from data_collection.
        spatial_subset : str, optional
            Spatial subset applied to data.
        cls : `~specutils.Spectrum1D`, `~astropy.nddata.CCDData`, optional
            The type that data will be returned as.

        Returns
        -------
        data : cls
            Data is returned as type cls with subsets applied.

        """
        return self._get_data(data_label=data_label, spatial_subset=spatial_subset, cls=cls)

    def get_ref_data(self):
        return get_reference_image_data(self.app)


def split_filename_with_fits_ext(filename):
    """Split a ``filename[ext]`` input into filename and FITS extension.

    Parameters
    ----------
    filename : str
        Can be a plain filename or ``filename[ext]``. The latter is a form
        of input that is commonly used by DS9. Example values:

        * ``'myimage.fits'``
        * ``'myimage.fits[SCI]'`` (assumes ``EXTVER=1``)
        * ``'myimage.fits[SCI,1]'``

    Returns
    -------
    filepath : str
        Path to the file, without extension.

    ext : str, tuple, or `None`
        FITS extension, if given. Examples: ``'SCI'`` or ``('SCI', 1)``

    data_label : str
        Human-readable data label for Glue. Extension info will be added
        later in the parser.

    """
    s = os.path.splitext(filename)
    ext_match = re.match(r'(.+)\[(.+)\]', s[1])
    if ext_match is None:
        sfx = s[1]
        ext = None
    else:
        sfx = ext_match.group(1)
        ext = ext_match.group(2)
        if ',' in ext:
            ext = ext.split(',')
            ext[1] = int(ext[1])
            ext = tuple(ext)
        elif not re.match(r'\D+', ext):
            ext = int(ext)

    filepath = f'{s[0]}{sfx}'
    data_label = os.path.basename(s[0])

    return filepath, ext, data_label



import warnings

import pytest
import numpy as np
from astropy import units as u
from astropy.io import fits
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.data import get_pkg_data_filename
from numpy.testing import assert_allclose, assert_array_equal
from photutils.aperture import (ApertureStats, CircularAperture, EllipticalAperture,
                                RectangularAperture, EllipticalAnnulus)
from photutils.datasets import make_4gaussians_image
from regions import (CircleAnnulusPixelRegion, CirclePixelRegion, EllipsePixelRegion,
                     RectanglePixelRegion, PixCoord)

from jdaviz.configs.imviz.plugins.aper_phot_simple.aper_phot_simple import (
    _curve_of_growth, _radial_profile)
from jdaviz.configs.imviz.tests.utils import BaseImviz_WCS_WCS, BaseImviz_WCS_NoWCS
from jdaviz.core.custom_units_and_equivs import PIX2


class TestSimpleAperPhot(BaseImviz_WCS_WCS):
    def test_plugin_wcs_dithered(self):
        self.imviz.link_data(align_by='wcs')  # They are dithered by 1 pixel on X

        reg = CirclePixelRegion(center=PixCoord(x=4.5, y=4.5), radius=4.5).to_sky(self.wcs_1)
        self.imviz.plugins['Subset Tools'].import_region(reg)

        phot_plugin = self.imviz.app.get_tray_item_from_name('imviz-aper-phot-simple')

        # Model fitting is already tested in astropy.
        # Here, we enable it just to make sure it does not crash.
        phot_plugin.fit_radial_profile = True

        # Make sure invalid Data/Subset selection does not crash plugin.
        with pytest.raises(ValueError):
            phot_plugin.dataset_selected = 'no_such_data'
        with pytest.raises(ValueError):
            # will raise an error and revert to first entry
            phot_plugin.aperture_selected = 'no_such_subset'
        assert phot_plugin.aperture_selected == ''
        phot_plugin.aperture_selected = phot_plugin.aperture.labels[0]
        assert_allclose(phot_plugin.background_value, 0)

        phot_plugin.dataset_selected = 'has_wcs_1[SCI,1]'
        phot_plugin.aperture_selected = phot_plugin.aperture.labels[0]
        with pytest.raises(ValueError):
            phot_plugin.background_selected = 'no_such_subset'
        assert phot_plugin.background_selected == 'Manual'
        assert_allclose(phot_plugin.background_value, 0)

        # Perform photometry on both images using same Subset.
        phot_plugin.dataset_selected = 'has_wcs_1[SCI,1]'
        phot_plugin.aperture_selected = 'Subset 1'
        assert phot_plugin.dataset.selected_dc_item is not None
        phot_plugin.vue_do_aper_phot()
        tbl = phot_plugin.export_table()
        assert len(tbl) == 1

        phot_plugin.dataset_selected = 'has_wcs_2[SCI,1]'
        phot_plugin.current_plot_type = 'Radial Profile (Raw)'
        assert phot_plugin.dataset.selected_dc_item is not None
        assert phot_plugin.aperture.selected_spatial_region is not None
        assert phot_plugin.background.labels == ['Manual', 'Subset 1']
        assert_allclose(phot_plugin.background_value, 0)
        assert_allclose(phot_plugin.counts_factor, 0)
        assert_allclose(phot_plugin.pixel_area, 0)
        assert_allclose(phot_plugin.flux_scaling, 0)
        phot_plugin.vue_do_aper_phot()
        tbl = phot_plugin.export_table()
        assert len(tbl) == 2
        assert phot_plugin.plot_available
        assert len(phot_plugin.plot.layers['profile'].layer.data['x']) > 0

        # Check photometry results.
        assert tbl.colnames == [
            'id', 'xcenter', 'ycenter', 'sky_center', 'background', 'sum',
            'sum_aper_area', 'pixarea_tot', 'aperture_sum_counts', 'aperture_sum_counts_err',
            'counts_fac', 'aperture_sum_mag', 'flux_scaling', 'min', 'max', 'mean', 'median',
            'mode', 'std', 'mad_std', 'var', 'biweight_location', 'biweight_midvariance',
            'fwhm', 'semimajor_sigma', 'semiminor_sigma', 'orientation', 'eccentricity',
            'data_label', 'subset_label', 'timestamp']
        assert_array_equal(tbl['id'], [1, 2])
        assert_allclose(tbl['background'], 0)
        assert_quantity_allclose(tbl['sum_aper_area'], [63.617251, 62.22684693104279] * PIX2, rtol=1e-4)  # noqa
        assert_array_equal(tbl['pixarea_tot'], None)
        assert_array_equal(tbl['aperture_sum_counts'], None)
        assert_array_equal(tbl['aperture_sum_counts_err'], None)
        assert_array_equal(tbl['counts_fac'], None)
        assert_array_equal(tbl['aperture_sum_mag'], None)
        assert_array_equal(tbl['flux_scaling'], None)
        assert_allclose(tbl['min'], 1)
        assert_allclose(tbl['max'], 1)
        assert_allclose(tbl['mean'], 1)
        assert_allclose(tbl['median'], 1)
        assert_allclose(tbl['mode'], 1)
        assert_allclose(tbl['std'], 0)
        assert_allclose(tbl['mad_std'], 0)
        assert_allclose(tbl['var'], 0)
        assert_allclose(tbl['biweight_location'], 1)
        assert_allclose(tbl['biweight_midvariance'], 0)
        assert_quantity_allclose(tbl['fwhm'], 5.15018758 * u.pix)
        assert_quantity_allclose(tbl['semimajor_sigma'], 2.18708329 * u.pix)
        assert_quantity_allclose(tbl['semiminor_sigma'], 2.18708329 * u.pix)
        assert_quantity_allclose(tbl['orientation'], 0 * u.deg)
        assert_quantity_allclose(tbl['eccentricity'], 0)
        assert_array_equal(tbl['data_label'], ['has_wcs_1[SCI,1]', 'has_wcs_2[SCI,1]'])
        assert_array_equal(tbl['subset_label'], ['Subset 1', 'Subset 1'])
        assert tbl['timestamp'].scale == 'utc'

        # Sky is the same but xcenter different due to dithering.
        # The aperture sum is different too because mask is a little off limit in second image.
        assert_quantity_allclose(tbl['xcenter'], [4.5, 5.5] * u.pix, rtol=1e-4)
        assert_quantity_allclose(tbl['ycenter'], 4.5 * u.pix, rtol=1e-4)
        sky = tbl['sky_center']
        assert_allclose(sky.ra.deg, 337.518943)
        assert_allclose(sky.dec.deg, -20.832083)
        assert_allclose(tbl['sum'], [63.617251, 62.22684693104279], rtol=1e-4)

        # Make sure it also works on an ellipse subset.
        reg = EllipsePixelRegion(center=PixCoord(x=4.5, y=2.0), width=9.0, height=4.0).to_sky(self.wcs_1)  # noqa: E501
        self.imviz.plugins['Subset Tools'].combination_mode = 'new'
        self.imviz.plugins['Subset Tools'].import_region(reg)

        phot_plugin.dataset_selected = 'has_wcs_1[SCI,1]'
        phot_plugin.aperture_selected = 'Subset 2'
        phot_plugin.current_plot_type = 'Radial Profile'
        phot_plugin.vue_do_aper_phot()
        tbl = phot_plugin.export_table()
        assert len(tbl) == 3  # New result is appended
        assert tbl[-1]['id'] == 3
        assert_quantity_allclose(tbl[-1]['xcenter'], 4.5 * u.pix, rtol=1e-4)
        assert_quantity_allclose(tbl[-1]['ycenter'], 2 * u.pix, rtol=1e-4)
        sky = tbl[-1]['sky_center']
        assert_allclose(sky.ra.deg, 337.51894336144454, rtol=1e-4)
        assert_allclose(sky.dec.deg, -20.832777499255897, rtol=1e-4)
        assert_quantity_allclose(tbl[-1]['sum_aper_area'], 28.274334 * PIX2, rtol=1e-4)
        assert_allclose(tbl[-1]['sum'], 28.274334, rtol=1e-4)
        assert_allclose(tbl[-1]['mean'], 1, rtol=1e-4)
        assert tbl[-1]['data_label'] == 'has_wcs_1[SCI,1]'
        assert tbl[-1]['subset_label'] == 'Subset 2'

        # Make sure it also works on a rectangle subset.
        # We also subtract off background from itself here.
        reg = RectanglePixelRegion(center=PixCoord(x=4.5, y=4.5), width=9, height=9).to_sky(self.wcs_1)  # noqa: E501
        self.imviz.plugins['Subset Tools'].combination_mode = 'new'
        self.imviz.plugins['Subset Tools'].import_region(reg)

        phot_plugin.dataset_selected = 'has_wcs_1[SCI,1]'
        phot_plugin.aperture_selected = 'Subset 3'
        phot_plugin.background_selected = 'Subset 3'
        assert_allclose(phot_plugin.background_value, 1)
        phot_plugin.vue_do_aper_phot()
        tbl = phot_plugin.export_table()
        assert len(tbl) == 4  # New result is appended
        assert tbl[-1]['id'] == 4
        assert_quantity_allclose(tbl[-1]['xcenter'], 4.5 * u.pix)
        assert_quantity_allclose(tbl[-1]['ycenter'], 4.5 * u.pix)
        sky = tbl[-1]['sky_center']
        assert_allclose(sky.ra.deg, 337.51894336144454, rtol=1e-4)
        assert_allclose(sky.dec.deg, -20.832083, rtol=1e-4)
        assert_quantity_allclose(tbl[-1]['sum_aper_area'], 81 * PIX2)
        assert_allclose(tbl[-1]['sum'], 0)
        assert_allclose(tbl[-1]['mean'], 0)
        assert tbl[-1]['data_label'] == 'has_wcs_1[SCI,1]'
        assert tbl[-1]['subset_label'] == 'Subset 3'

        # Make sure background auto-updates.
        phot_plugin.background_selected = 'Manual'
        assert_allclose(phot_plugin.background_value, 1)  # Keeps last value
        phot_plugin.background_selected = 'Subset 1'
        assert_allclose(phot_plugin.background_value, 1)

        hdu3 = fits.ImageHDU(np.ones((10, 10)) + 1, name='SCI')
        hdu3.header.update(self.wcs_2.to_header())
        self.imviz.load_data(hdu3, data_label='twos')
        phot_plugin.dataset_selected = 'twos[SCI,1]'
        assert_allclose(phot_plugin.background_value, 2)  # Recalculate based on new Data

        # Curve of growth
        phot_plugin.current_plot_type = 'Curve of Growth'
        phot_plugin.vue_do_aper_phot()
        assert phot_plugin.plot.figure.title == 'Curve of growth from aperture center'

    def test_batch_unpack(self):
        # TODO: remove ._obj when API is made public
        phot_plugin = self.imviz.plugins['Aperture Photometry']._obj

        # NOTE: these input values are not currently validated, so it does not matter that the
        # datasets and subsets do not exist with these names (if that changes, this test will
        # need to be udpated accordingly)
        unpacked = phot_plugin.unpack_batch_options(dataset=['image1', 'image2'],
                                                    aperture=['Subset 1', 'Subset 2'],
                                                    background=['Subset 3'],
                                                    flux_scaling=3)
        assert unpacked == [{'aperture': 'Subset 1',
                             'dataset': 'image1',
                             'background': 'Subset 3',
                             'flux_scaling': 3},
                            {'aperture': 'Subset 2',
                             'dataset': 'image1',
                             'background': 'Subset 3',
                             'flux_scaling': 3},
                            {'aperture': 'Subset 1',
                             'dataset': 'image2',
                             'background': 'Subset 3',
                             'flux_scaling': 3},
                            {'aperture': 'Subset 2',
                             'dataset': 'image2',
                             'background': 'Subset 3',
                             'flux_scaling': 3}]

    def test_batch_phot(self):
        self.imviz.link_data(align_by='wcs')  # They are dithered by 1 pixel on X
        self.imviz.plugins['Subset Tools'].import_region(
            CirclePixelRegion(center=PixCoord(x=4.5, y=4.5), radius=4.5)
        )  # Draw a circle

        # TODO: remove ._obj when API is made public
        phot_plugin = self.imviz.plugins['Aperture Photometry']._obj
        assert phot_plugin.dataset.choices == ['has_wcs_1[SCI,1]', 'has_wcs_2[SCI,1]']
        assert phot_plugin.aperture.choices == ['Subset 1']

        phot_plugin.aperture.selected = 'Subset 1'
        phot_plugin.calculate_batch_photometry([{'dataset': 'has_wcs_1[SCI,1]', 'aperture': 'Subset 1'},  # noqa
                                                {'dataset': 'has_wcs_2[SCI,1]'}])

        assert len(phot_plugin.table) == 2

        with pytest.raises(RuntimeError):
            phot_plugin.calculate_batch_photometry([{'dataset': 'has_wcs_1[SCI,1]', 'aperture': 'DNE'},  # noqa
                                                    {'dataset': 'has_wcs_2[SCI,1]', 'aperture': 'Subset 1'}])  # noqa

        # second entry above should have been successful, resulting in one addition to the results
        assert len(phot_plugin.table) == 3

        # now run through the UI directly
        phot_plugin.multiselect = True
        phot_plugin.dataset.select_all()
        phot_plugin.aperture.select_none()
        assert len(phot_plugin.unpack_batch_options()) == 0
        phot_plugin.vue_do_aper_phot()

        assert len(phot_plugin.table) == 3
        phot_plugin.aperture.select_all()
        assert len(phot_plugin.unpack_batch_options()) == 2
        phot_plugin.vue_do_aper_phot()
        assert len(phot_plugin.table) == 5


class TestSimpleAperPhot_NoWCS(BaseImviz_WCS_NoWCS):
    def test_plugin_no_wcs(self):
        # Most things already tested above, so not re-tested here.
        self.imviz.plugins['Subset Tools'].import_region(
            CirclePixelRegion(center=PixCoord(x=4.5, y=4.5), radius=4.5)
        )  # Draw a circle
        phot_plugin = self.imviz.app.get_tray_item_from_name('imviz-aper-phot-simple')

        phot_plugin.dataset_selected = 'has_wcs[SCI,1]'
        phot_plugin.aperture_selected = 'Subset 1'
        phot_plugin.vue_do_aper_phot()
        tbl = phot_plugin.export_table()
        assert len(tbl) == 1

        phot_plugin.dataset_selected = 'no_wcs[SCI,1]'
        phot_plugin.vue_do_aper_phot()
        tbl = phot_plugin.export_table()
        assert len(tbl) == 1  # Old table discarded due to incompatible column
        assert_array_equal(tbl['sky_center'], None)


class TestAdvancedAperPhot:
    @pytest.fixture(autouse=True)
    def setup_class(self, imviz_helper):
        # Reference image
        fn_1 = get_pkg_data_filename('data/gauss100_fits_wcs.fits')
        imviz_helper.load_data(fn_1)
        # Different pixel scale
        imviz_helper.load_data(get_pkg_data_filename('data/gauss100_fits_wcs_block_reduced.fits'))
        # Different pixel scale + rotated
        imviz_helper.load_data(get_pkg_data_filename('data/gauss100_fits_wcs_block_reduced_rotated.fits'))  # noqa: E501

        # Link them by WCS
        imviz_helper.link_data(align_by='wcs')
        w = imviz_helper.app.data_collection[0].coords

        # Regions to be used for aperture photometry
        imviz_helper.plugins['Subset Tools'].import_region([
            CirclePixelRegion(center=PixCoord(x=145.1, y=168.3), radius=5).to_sky(w),
            CirclePixelRegion(center=PixCoord(x=48.3, y=200.3), radius=5).to_sky(w),
            EllipsePixelRegion(center=PixCoord(x=84.7, y=224.1), width=23, height=9, angle=2.356 * u.rad).to_sky(w),  # noqa: E501
            RectanglePixelRegion(center=PixCoord(x=229, y=152), width=17, height=7).to_sky(w)],
            combination_mode='new')

        self.imviz = imviz_helper
        self.viewer = imviz_helper.default_viewer._obj
        self.phot_plugin = imviz_helper.plugins["Aperture Photometry"]._obj

    @pytest.mark.parametrize(('data_label', 'local_bkg'), [
        ('gauss100_fits_wcs[PRIMARY,1]', 5.0),
        ('gauss100_fits_wcs_block_reduced[PRIMARY,1]', 20.0),
        ('gauss100_fits_wcs_block_reduced_rotated[PRIMARY,1]', 20.0)])
    @pytest.mark.parametrize(('subset_label', 'expected_sum'), [
        ('Subset 1', 738.8803424408962),
        ('Subset 2', 857.5194857987592),
        ('Subset 3', 472.17364321556005),
        ('Subset 4', 837.0023608207703)])
    def test_aperphot(self, data_label, local_bkg, subset_label, expected_sum):
        """All data should give similar result for the same Subset."""
        self.phot_plugin.dataset_selected = data_label
        self.phot_plugin.aperture_selected = subset_label
        self.phot_plugin.background_selected = 'Manual'
        self.phot_plugin.background_value = local_bkg
        self.phot_plugin.vue_do_aper_phot()
        tbl = self.phot_plugin.export_table()

        # Imperfect down-sampling and imperfect apertures, so 10% is good enough.
        assert_allclose(tbl['sum'][-1], expected_sum, rtol=0.1)

    @pytest.mark.parametrize(('data_label', 'fac'), [
        ('gauss100_fits_wcs[PRIMARY,1]', 1),
        ('gauss100_fits_wcs_block_reduced[PRIMARY,1]', 4),
        ('gauss100_fits_wcs_block_reduced_rotated[PRIMARY,1]', 4)
    ])
    @pytest.mark.parametrize(('bg_label', 'expected_bg'), [
        ('Subset 2', 12.269274711608887),
        ('Subset 3', 7.935906410217285),
        ('Subset 4', 11.120951652526855)])
    def test_sky_background(self, data_label, fac, bg_label, expected_bg):
        """All background (median) should give similar result for the same Subset.
        Down-sampled data has higher factor due to flux conservation.
        """
        self.phot_plugin.dataset_selected = data_label
        self.phot_plugin.aperture_selected = "Subset 1"  # Does not matter
        self.phot_plugin.background_selected = bg_label

        # Imperfect down-sampling and abusing apertures, so 10% is good enough.
        assert_allclose(float(self.phot_plugin.background_value), expected_bg * fac, rtol=0.1)


def test_annulus_background(imviz_helper):
    gauss4 = make_4gaussians_image()  # The background has a mean of 5 with noise
    ones = np.ones(gauss4.shape)
    bg_4gauss_1 = 5.802287
    bg_4gauss_2 = 5.052332
    bg_4gauss_3 = 45.416834
    bg_4gauss_4 = 4.939397

    imviz_helper.load_data(gauss4, data_label='four_gaussians')
    imviz_helper.load_data(ones, data_label='ones')

    phot_plugin = imviz_helper.app.get_tray_item_from_name('imviz-aper-phot-simple')
    phot_plugin.dataset_selected = 'ones'

    # Mark an object of interest
    circle_1 = CirclePixelRegion(center=PixCoord(x=150, y=25), radius=7)
    # Load annulus (this used to be part of the plugin but no longer)
    annulus_1 = CircleAnnulusPixelRegion(
        PixCoord(x=150, y=25), inner_radius=7, outer_radius=17)
    imviz_helper.plugins['Subset Tools'].import_region([circle_1, annulus_1],
                                                       combination_mode='new')

    phot_plugin.aperture_selected = 'Subset 1'
    phot_plugin.background_selected = 'Subset 2'

    # Check annulus for ones
    assert_allclose(phot_plugin.background_value, 1)

    # Switch data
    phot_plugin.dataset_selected = 'four_gaussians'
    assert_allclose(phot_plugin.background_value, bg_4gauss_1)  # Changed

    # Draw ellipse on another object
    ellipse_1 = EllipsePixelRegion(center=PixCoord(x=20.5, y=37.5), width=41, height=15)
    # Load annulus (this used to be part of the plugin but no longer)
    annulus_2 = CircleAnnulusPixelRegion(
        PixCoord(x=20.5, y=37.5), inner_radius=20.5, outer_radius=30.5)
    imviz_helper.plugins['Subset Tools'].import_region([ellipse_1, annulus_2],
                                                       combination_mode='new')

    # Subset 4 (annulus) should be available in both sets of choices, but invalid for selection as
    # aperture
    assert 'Subset 4' in phot_plugin.aperture.choices
    assert 'Subset 4' in phot_plugin.background.choices

    phot_plugin.aperture_selected = 'Subset 4'
    assert not phot_plugin.aperture.selected_validity.get('is_aperture', True)
    with pytest.raises(ValueError, match="Selected aperture is not valid"):
        phot_plugin.calculate_photometry()

    phot_plugin.aperture_selected = 'Subset 3'
    assert phot_plugin.aperture.selected_validity.get('is_aperture', False)
    phot_plugin.background_selected = 'Subset 4'

    # Check new annulus for four_gaussians
    assert_allclose(phot_plugin.background_value, bg_4gauss_2)  # Changed

    # Switch to manual, should not change
    phot_plugin.background_selected = 'Manual'
    assert_allclose(phot_plugin.background_value, bg_4gauss_2)

    # Switch to Subset, should change a lot because this is not background area
    phot_plugin.background_selected = 'Subset 1'
    assert_allclose(phot_plugin.background_value, bg_4gauss_3)

    # Switch back to annulus, should be same as before in same mode
    phot_plugin.background_selected = 'Subset 4'
    assert_allclose(phot_plugin.background_value, bg_4gauss_2)

    # Edit the annulus and make sure background updates
    subset_plugin = imviz_helper.plugins['Subset Tools']._obj
    subset_plugin.subset_selected = "Subset 4"
    subset_plugin._set_value_in_subset_definition(0, "X Center (pixels)", "value", 25.5)
    subset_plugin._set_value_in_subset_definition(0, "Y Center (pixels)", "value", 42.5)
    subset_plugin._set_value_in_subset_definition(0, "Inner Radius (pixels)", "value", 40)
    subset_plugin._set_value_in_subset_definition(0, "Outer Radius (pixels)", "value", 45)
    subset_plugin.vue_update_subset()
    assert_allclose(phot_plugin.background_value, bg_4gauss_4)


def test_fit_radial_profile_with_nan(imviz_helper):
    gauss4 = make_4gaussians_image()  # The background has a mean of 5 with noise
    # Insert NaN
    gauss4[25, 150] = np.nan

    imviz_helper.load_data(gauss4, data_label='four_gaussians')

    # Mark an object of interest
    circle_1 = CirclePixelRegion(center=PixCoord(x=150, y=25), radius=7)
    imviz_helper.plugins['Subset Tools'].import_region(
        [circle_1], combination_mode='new')

    phot_plugin = imviz_helper.app.get_tray_item_from_name('imviz-aper-phot-simple')
    phot_plugin.dataset_selected = 'four_gaussians'
    phot_plugin.aperture_selected = 'Subset 1'
    phot_plugin.current_plot_type = 'Radial Profile'
    phot_plugin.fit_radial_profile = True
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")  # Fitter warnings do not matter, only want error.
        phot_plugin.vue_do_aper_phot()
    tbl = imviz_helper.plugins['Aperture Photometry'].export_table()

    assert phot_plugin.result_failed_msg == ''
    assert_allclose(tbl['sum'][0], 8590.419296)


# NOTE: Extracting the cutout for radial profile is aperture
#       shape agnostic, so we use ellipse as representative case.
# NOTE: This test only tests the radial profile algorithm and does
#       not care if the actual plugin use centroid or not.
class TestRadialProfile():
    def setup_class(self):
        data = np.ones((51, 51)) * u.nJy
        aperture = EllipticalAperture((25, 25), 20, 15)
        phot_aperstats = ApertureStats(data, aperture)
        self.data_cutout = phot_aperstats.data_cutout
        self.bbox = phot_aperstats.bbox
        self.centroid = phot_aperstats.centroid

    def test_profile_raw(self):
        x_arr, y_arr = _radial_profile(self.data_cutout, self.bbox, self.centroid, raw=True)
        # Too many data points to compare each one for X.
        assert x_arr.shape == y_arr.shape == (923, )
        assert_allclose(x_arr.min(), 0)
        assert_allclose(x_arr.max(), 19.4164878389476)
        assert_allclose(y_arr, 1)

    def test_profile_imexam(self):
        x_arr, y_arr = _radial_profile(self.data_cutout, self.bbox, self.centroid, raw=False)
        assert_allclose(x_arr, range(20))
        assert_allclose(y_arr, 1)


# NOTE: This test only tests the curve of growth algorithm and does
#       not care if the actual plugin use centroid or not.
@pytest.mark.parametrize('with_unit', (False, True))
def test_curve_of_growth(with_unit):
    data = np.ones((51, 51))
    cen = (25, 25)
    if with_unit:
        data = data << (u.MJy / u.sr)
        bg = 0 * data.unit
        pixarea_fac = 1 * u.sr
        expected_ylabel = 'MJy'
    else:
        bg = 0
        pixarea_fac = None
        expected_ylabel = 'Value'

    apertures = (CircularAperture(cen, 20),
                 EllipticalAperture(cen, 20, 15),
                 RectangularAperture(cen, 20, 15))

    for aperture in apertures:
        astat = ApertureStats(data, aperture)
        final_sum = astat.sum
        if pixarea_fac is not None:
            final_sum = final_sum * pixarea_fac
        x_arr, sum_arr, x_label, y_label = _curve_of_growth(
            data, astat.centroid, aperture, final_sum, background=bg, pixarea_fac=pixarea_fac)
        assert_allclose(x_arr, [2, 4, 6, 8, 10, 12, 14, 16, 18, 20])
        assert y_label == expected_ylabel

        if isinstance(aperture, CircularAperture):
            assert x_label == 'Radius (pix)'
            assert_allclose(sum_arr, [
                12.566371, 50.265482, 113.097336, 201.06193, 314.159265,
                452.389342, 615.75216, 804.247719, 1017.87602, 1256.637061])
        elif isinstance(aperture, EllipticalAperture):
            assert x_label == 'Semimajor axis (pix)'
            assert_allclose(sum_arr, [
                9.424778, 37.699112, 84.823002, 150.796447, 235.619449,
                339.292007, 461.81412, 603.185789, 763.407015, 942.477796])
        else:  # RectangularAperture
            assert x_label == 'Width (pix)'
            assert_allclose(sum_arr, [3, 12, 27, 48, 75, 108, 147, 192, 243, 300])

    with pytest.raises(TypeError, match='Unsupported aperture'):
        _curve_of_growth(data, cen, EllipticalAnnulus(cen, 3, 8, 5), 100,
                         pixarea_fac=pixarea_fac)


def test_cubeviz_batch(cubeviz_helper, spectrum1d_cube_fluxunit_jy_per_steradian):
    cubeviz_helper.load_data(spectrum1d_cube_fluxunit_jy_per_steradian, data_label='test')
    phot_plugin = cubeviz_helper.plugins['Aperture Photometry']._obj
    uc_plugin = cubeviz_helper.plugins['Unit Conversion']
    subset_plugin = cubeviz_helper.plugins['Subset Tools']

    subset_plugin.import_region(CirclePixelRegion(center=PixCoord(x=5, y=5), radius=2),
                                combination_mode='new')
    subset_plugin.import_region(CirclePixelRegion(center=PixCoord(x=3, y=3), radius=2),
                                combination_mode='new')

    phot_plugin.dataset_selected = 'test[FLUX]'
    phot_plugin.multiselect = True
    phot_plugin.aperture.selected = ['Subset 1', 'Subset 2']

    phot_plugin.calculate_batch_photometry(full_exceptions=True)
    assert len(phot_plugin.table) == 2
    tbl = cubeviz_helper.plugins['Aperture Photometry'].export_table()
    assert_quantity_allclose(tbl['sum'], [5.980836e-12, 2.037396e-10] * u.Jy, rtol=1e-4)

    # Test that it still works after unit conversion
    uc_plugin.flux_unit = 'MJy'

    phot_plugin.calculate_batch_photometry(full_exceptions=True)

    assert len(phot_plugin.table) == 4
    tbl = cubeviz_helper.plugins['Aperture Photometry'].export_table()
    # get_aperture_photometry_results converts all to the same units
    assert_quantity_allclose(tbl['sum'],
                             [5.980836e-12, 2.037396e-10, 5.980836e-12, 2.037396e-10] * u.Jy,
                             rtol=1e-4)



import gwcs
import numpy as np
import pytest
from astropy import units as u
from astropy.coordinates import ICRS
from astropy.modeling import models
from astropy.wcs import WCS
from gwcs import coordinate_frames as cf
from numpy.testing import assert_allclose

from jdaviz.configs.imviz import wcs_utils
from jdaviz.configs.imviz.plugins.orientation.orientation import base_wcs_layer_label
from jdaviz.configs.imviz.tests.utils import BaseImviz_WCS_GWCS, create_example_gwcs


def test_simple_fits_wcs():
    # CTYPE : ''  ''
    # CRVAL : 0.0  0.0
    # CRPIX : 0.0  0.0
    # PC1_1 PC1_2  : 1.0  0.0
    # PC2_1 PC2_2  : 0.0  1.0
    # CDELT : 1.0  1.0
    # NAXIS : 0  0
    w = WCS()
    result = wcs_utils.get_compass_info(w, (100, 100), r_fac=0.25)
    assert_allclose(result[:-1], (50, 50,
                                  50, 73.573199,
                                  73.573199, 47.538975,
                                  0, 95.960048))
    assert result[-1]

    # https://learn.astropy.org/tutorials/celestial_coords1.html
    w = WCS({'CTYPE1': 'RA---TAN',
             'CUNIT1': 'deg',
             'CDELT1': -0.0002777777778,
             'CRPIX1': 1,
             'CRVAL1': 337.5202808,
             'NAXIS1': 1024,
             'CTYPE2': 'DEC--TAN',
             'CUNIT2': 'deg',
             'CDELT2': 0.0002777777778,
             'CRPIX2': 1,
             'CRVAL2': -20.833333059999998,
             'NAXIS2': 1024})
    result = wcs_utils.get_compass_info(w, (1024, 1024), r_fac=0.25)
    assert_allclose(result[:-1], (512.0, 512.0,
                                  512.2416, 768.0007,
                                  256.0009, 512.24,
                                  0.05406877, -90.00035))
    assert not result[-1]


def test_hst_acs_flt_wcs():
    # jb5g05ubq_flt.fits -- has distortion
    w = WCS({'WCSAXES': 2,
             'CRPIX1': 2100.0,
             'CRPIX2': 1024.0,
             'PC1_1': -1.14852e-05,
             'PC1_2': 7.01477e-06,
             'PC2_1': 7.75765e-06,
             'PC2_2': 1.20927e-05,
             'CDELT1': 1.0,
             'CDELT2': 1.0,
             'CUNIT1': 'deg',
             'CUNIT2': 'deg',
             'CTYPE1': 'RA---TAN',
             'CTYPE2': 'DEC--TAN',
             'CRVAL1': 3.581704851882,
             'CRVAL2': -30.39197867265,
             'LONPOLE': 180.0,
             'LATPOLE': -30.39197867265,
             'MJDREF': 0.0,
             'RADESYS': 'ICRS'})
    result = wcs_utils.get_compass_info(w, (2048, 4096), r_fac=0.25)
    assert_allclose(result[:-1], (2048.0, 1024.0,
                                  2314.86981434716, 1460.9491537866957,
                                  1617.0522515837179, 1300.4560437373216,
                                  31.414796751823395, -88.73433642655598))
    assert not result[-1]


def test_simple_gwcs():
    # https://gwcs.readthedocs.io/en/latest/#getting-started
    shift_by_crpix = models.Shift(-(2048 - 1) * u.pix) & models.Shift(-(1024 - 1) * u.pix)
    matrix = np.array([[1.290551569736E-05, 5.9525007864732E-06],
                       [5.0226382102765E-06, -1.2644844123757E-05]])
    rotation = models.AffineTransformation2D(matrix * u.deg, translation=[0, 0] * u.deg)
    rotation.input_units_equivalencies = {"x": u.pixel_scale(1 * (u.deg / u.pix)),
                                          "y": u.pixel_scale(1 * (u.deg / u.pix))}
    rotation.inverse = models.AffineTransformation2D(np.linalg.inv(matrix) * u.pix,
                                                     translation=[0, 0] * u.pix)
    rotation.inverse.input_units_equivalencies = {"x": u.pixel_scale(1 * (u.pix / u.deg)),
                                                  "y": u.pixel_scale(1 * (u.pix / u.deg))}
    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(
        5.63056810618 * u.deg, -72.05457184279 * u.deg, 180 * u.deg)
    det2sky = shift_by_crpix | rotation | tan | celestial_rotation
    det2sky.name = "linear_transform"
    detector_frame = cf.Frame2D(name="detector", axes_names=("x", "y"), unit=(u.pix, u.pix))
    sky_frame = cf.CelestialFrame(reference_frame=ICRS(), name='icrs', unit=(u.deg, u.deg))
    pipeline = [(detector_frame, det2sky), (sky_frame, None)]
    w = gwcs.wcs.WCS(pipeline)

    result = wcs_utils.get_compass_info(w, (1024, 2048), r_fac=0.25)
    assert_allclose(result[:-1], (1024.0, 512.0,
                                  1131.026544, 279.446094,
                                  1262.004542, 606.285924,
                                  155.287048, -86.898131))
    assert not result[-1]


# TODO: Add more tests.
class TestWCSOnly(BaseImviz_WCS_GWCS):

    # TODO: Replace private API calls with more public ones when available.
    def test_non_wcs_layer_labels(self):
        from jdaviz.utils import get_wcs_only_layer_labels

        self.imviz.link_data(align_by="wcs")
        assert len(self.imviz.app.data_collection) == 3

        # Confirm the WCS-only layer is created by WCS-linking .
        assert len(get_wcs_only_layer_labels(self.imviz.app)) == 1

        # Load a WCS-only layer, bypassing normal labeling scheme.
        ndd = wcs_utils._get_rotated_nddata_from_label(
            app=self.imviz.app,
            data_label="fits_wcs[DATA]",
            rotation_angle=5 * u.deg
        )
        self.imviz.load_data(ndd, data_label='ndd')
        assert self.imviz.app.data_collection[3].label == 'ndd'

        # Confirm that all data in collection are labeled.
        assert len(self.imviz.app.state.layer_icons) == 4  # 3 + 1

        # Confirm the new WCS-only layer is logged.
        assert len(get_wcs_only_layer_labels(self.imviz.app)) == 2

        # Load a second WCS-only layer.
        ndd2 = wcs_utils._get_rotated_nddata_from_label(
            app=self.imviz.app,
            data_label="fits_wcs[DATA]",
            rotation_angle=45 * u.deg
        )
        self.imviz.load_data(ndd2, data_label="rot: 45.00 deg")
        assert self.imviz.app.data_collection[4].label == "rot: 45.00 deg"

        # Confirm that all data in collection are labeled.
        assert len(self.imviz.app.data_collection) == 5  # 3 + 2
        assert len(self.imviz.app.state.layer_icons) == 5

        # Confirm the second WCS-only layer is logged
        assert len(get_wcs_only_layer_labels(self.imviz.app)) == 3

        # First entry is image data and the default reference data.
        assert self.imviz.app.state.layer_icons["fits_wcs[DATA]"] == "a"
        assert self.viewer.state.reference_data.label == base_wcs_layer_label


def test_get_rotated_nddata_from_label_no_wcs(imviz_helper):
    a = np.zeros((2, 2), dtype=np.int8)
    imviz_helper.load_data(a, data_label="no_wcs")
    with pytest.raises(ValueError, match=r".*has no WCS for rotation"):
        wcs_utils._get_rotated_nddata_from_label(imviz_helper.app, "no_wcs", 0 * u.deg)


def test_compute_scale():
    # NOTE: compute_scale does not work with FITS WCS.
    image_gwcs = create_example_gwcs((10, 10))
    fiducial = [197.8925, -1.36555556] * u.deg

    pixscale = wcs_utils.compute_scale(image_gwcs, fiducial, None)
    assert_allclose(pixscale, 3.0555555555555554e-05)

    pixscale = wcs_utils.compute_scale(image_gwcs, fiducial, None, pscale_ratio=1e+5)
    assert_allclose(pixscale, 3.0555555555555554)



import os

from glue.config import viewer_tool
from astropy import units as u

from jdaviz.configs.mosviz.plugins.viewers import MosvizProfileView, MosvizProfile2DView
from jdaviz.core.tools import _MatchedZoomMixin, HomeZoom, BoxZoom, XRangeZoom, PanZoom, PanZoomX

__all__ = []

ICON_DIR = os.path.join(os.path.dirname(__file__), '..', '..', '..', 'data', 'icons')


class _MatchedXZoomMixin(_MatchedZoomMixin):
    match_axes = ('x',)
    disable_matched_zoom_in_other_viewer = True

    def _is_matched_viewer(self, viewer):
        return isinstance(viewer, (MosvizProfile2DView, MosvizProfileView))

    def _map_limits(self, from_viewer, to_viewer, limits={}):
        components = self.viewer.state.data_collection[0]._components
        # Determine cid for spectral axis
        for key in components.keys():
            strkey = str(key)
            if 'Wavelength' in strkey or 'Wave' in strkey:
                native_unit = u.Unit(self.viewer.state.data_collection[0].get_component(strkey).units)  # noqa
                break
        else:
            # no matches found
            native_unit = ''
        current_display_unit = u.Unit(self.viewer.jdaviz_helper.app._get_display_unit('spectral'))

        if isinstance(from_viewer, MosvizProfileView) and isinstance(to_viewer, MosvizProfile2DView):  # noqa
            if native_unit != current_display_unit and native_unit != '':
                limits['x_min'] = (limits['x_min'] * native_unit).to_value(
                                   current_display_unit, equivalencies=u.spectral()
                                )
                limits['x_max'] = (limits['x_max'] * native_unit).to_value(
                                   current_display_unit, equivalencies=u.spectral()
                                   )
            limits['x_min'], limits['x_max'] = to_viewer.world_to_pixel_limits((limits['x_min'],
                                                                                limits['x_max']))
        elif isinstance(from_viewer, MosvizProfile2DView) and isinstance(to_viewer, MosvizProfileView):  # noqa
            limits['x_min'], limits['x_max'] = from_viewer.pixel_to_world_limits((limits['x_min'],
                                                                                  limits['x_max']))
            if native_unit != current_display_unit and native_unit != '':
                limits['x_min'] = (limits['x_min'] * native_unit).to_value(
                                   current_display_unit, equivalencies=u.spectral()
                                   )
                limits['x_max'] = (limits['x_max'] * native_unit).to_value(
                                   current_display_unit, equivalencies=u.spectral()
                                   )
        return limits


@viewer_tool
class MosvizHomeZoom(_MatchedXZoomMixin, HomeZoom):
    icon = os.path.join(ICON_DIR, 'home_match.svg')
    tool_id = 'mosviz:homezoom'
    action_text = 'Reset zoom'
    tool_tip = 'Reset zoom to show all visible data, matching x-limits in all viewers'


@viewer_tool
class MosvizBoxZoom(_MatchedXZoomMixin, BoxZoom):
    icon = os.path.join(ICON_DIR, 'zoom_box_match.svg')
    tool_id = 'mosviz:boxzoom'
    action_text = 'Box zoom, matching x-limits between viewers'
    tool_tip = 'Zoom to a drawn rectangle, matching x-limits in all viewers'


@viewer_tool
class MosvizXRangeZoom(_MatchedXZoomMixin, XRangeZoom):
    icon = os.path.join(ICON_DIR, 'zoom_xrange_match.svg')
    tool_id = 'mosviz:xrangezoom'
    action_text = 'Horizontal zoom, matching x-limits between viewers'
    tool_tip = 'Zoom to a drawn horizontal region, matching x-limits in all viewers'


@viewer_tool
class MosvizPanZoom(_MatchedXZoomMixin, PanZoom):
    icon = os.path.join(ICON_DIR, 'panzoom_match.svg')
    tool_id = 'mosviz:panzoom'
    action_text = 'Pan, matching x-limits between viewers'
    tool_tip = 'Pan (click-drag) or zoom (scroll), matching x-limits in all viewers'  # noqa


@viewer_tool
class MosvizPanZoomX(_MatchedXZoomMixin, PanZoomX):
    icon = os.path.join(ICON_DIR, 'pan_x_match.svg')
    tool_id = 'mosviz:panzoom_x'
    action_text = 'Pan, matching x-limits between viewers'
    tool_tip = 'Pan (click-drag) or zoom (scroll) in x only, matching limits in all viewers'  # noqa



import numpy as np
import pytest
from astropy import units as u
from astropy.wcs import WCS
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.data import get_pkg_data_filename
from glue.core.roi import CircularROI, CircularAnnulusROI, EllipticalROI, RectangularROI, XRangeROI

from glue.core.subset_group import GroupedSubset
from regions import (PixCoord, CirclePixelRegion, CircleSkyRegion, RectanglePixelRegion,
                     EllipsePixelRegion, CircleAnnulusPixelRegion)
from numpy.testing import assert_allclose
from specutils import SpectralRegion, Spectrum1D
from astropy.nddata import NDData

from jdaviz.utils import get_subset_type, MultiMaskSubsetState


def test_region_from_subset_2d(cubeviz_helper):
    cubeviz_helper.load_data(np.ones((128, 128, 1)), data_label='Test 2D Flux')

    subset_plugin = cubeviz_helper.plugins['Subset Tools']

    cubeviz_helper.app.add_data_to_viewer('flux-viewer', 'Test 2D Flux')

    subset_plugin.import_region(
        EllipsePixelRegion(center=PixCoord(x=1, y=3.5), width=2.4, height=6.6))

    subsets = cubeviz_helper.app.get_subsets()
    reg = subsets.get('Subset 1')[0]['region']

    assert len(subsets) == 1
    assert isinstance(reg, EllipsePixelRegion)

    assert_allclose(reg.center.x, 1)
    assert_allclose(reg.center.y, 3.5)
    assert_allclose(reg.width, 2.4)
    assert_allclose(reg.height, 6.6)
    assert_allclose(reg.angle.value, 0)

    assert subset_plugin.subset == "Subset 1"
    assert subset_plugin._obj.subset_types == ["EllipticalROI"]
    assert subset_plugin._obj.is_centerable
    for key in ("orig", "value"):
        assert subset_plugin._obj._get_value_from_subset_definition(
            0, "X Center (pixels)", key) == 1
        assert subset_plugin._obj._get_value_from_subset_definition(
            0, "Y Center (pixels)", key) == 3.5
        assert subset_plugin._obj._get_value_from_subset_definition(
            0, "X Radius (pixels)", key) == 1.2
        assert subset_plugin._obj._get_value_from_subset_definition(
            0, "Y Radius (pixels)", key) == 3.3
        assert subset_plugin._obj._get_value_from_subset_definition(0, "Angle", key) == 0

    # Recenter GUI should not be exposed, but API call would raise exception.
    with pytest.raises(NotImplementedError, match='Cannot recenter'):
        subset_plugin.recenter()


def test_region_from_subset_3d(cubeviz_helper):
    cubeviz_helper.load_data(np.ones((128, 128, 256)), data_label='Test 3D Flux')

    subset_plugin = cubeviz_helper.plugins['Subset Tools']
    assert subset_plugin._obj.subset_selected == "Create New"

    cubeviz_helper.app.add_data_to_viewer('flux-viewer', 'Test 3D Flux')

    subset_plugin.import_region(RectangularROI(1, 3.5, -0.2, 3.3))

    subsets = cubeviz_helper.app.get_subsets()
    reg = cubeviz_helper.app.get_subsets('Subset 1', object_only=True)[0]

    assert len(subsets) == 1
    assert isinstance(reg, RectanglePixelRegion)

    assert_allclose(reg.center.x, 2.25)
    assert_allclose(reg.center.y, 1.55)
    assert_allclose(reg.width, 2.5)
    assert_allclose(reg.height, 3.5)
    assert_allclose(reg.angle.value, 0)

    assert subset_plugin._obj.subset_selected == "Subset 1"
    assert subset_plugin._obj.subset_types == ["RectangularROI"]
    assert subset_plugin._obj.is_centerable
    assert subset_plugin.get_center() == (2.25, 1.55)
    for key in ("orig", "value"):
        assert subset_plugin._obj._get_value_from_subset_definition(0, "Xmin (pixels)", key) == 1
        assert subset_plugin._obj._get_value_from_subset_definition(0, "Xmax (pixels)", key) == 3.5
        assert subset_plugin._obj._get_value_from_subset_definition(0, "Ymin (pixels)", key) == -0.2
        assert subset_plugin._obj._get_value_from_subset_definition(0, "Ymax (pixels)", key) == 3.3
        assert subset_plugin._obj._get_value_from_subset_definition(0, "Angle", key) == 0

    # Mimic user changing something in Subset GUI.
    subset_plugin._obj._set_value_in_subset_definition(0, "Xmin (pixels)", "value", 2)
    subset_plugin._obj._set_value_in_subset_definition(0, "Ymin (pixels)", "value", 0)
    subset_plugin._obj._set_value_in_subset_definition(0, "Angle", "value", 45)  # ccw deg
    # "orig" is unchanged until user clicks Update button.
    assert subset_plugin._obj._get_value_from_subset_definition(0, "Xmin (pixels)", "orig") == 1
    assert subset_plugin._obj._get_value_from_subset_definition(0, "Ymin (pixels)", "orig") == -0.2
    assert subset_plugin._obj._get_value_from_subset_definition(0, "Angle", "orig") == 0
    subset_plugin._obj.vue_update_subset()
    for key in ("orig", "value"):
        assert subset_plugin._obj._get_value_from_subset_definition(0, "Xmin (pixels)", key) == 2
        assert subset_plugin._obj._get_value_from_subset_definition(0, "Xmax (pixels)", key) == 3.5
        assert subset_plugin._obj._get_value_from_subset_definition(0, "Ymin (pixels)", key) == 0
        assert subset_plugin._obj._get_value_from_subset_definition(0, "Ymax (pixels)", key) == 3.3
        assert subset_plugin._obj._get_value_from_subset_definition(0, "Angle", key) == 45

    subsets = cubeviz_helper.app.get_subsets()
    reg = subsets.get('Subset 1')[0]['region']

    assert_allclose(reg.center.x, 2.75)
    assert_allclose(reg.center.y, 1.65)
    assert_allclose(reg.width, 1.5)
    assert_allclose(reg.height, 3.3)
    assert_allclose(reg.angle.to_value(u.deg), 45)  # Might be stored in radians

    # Move the rectangle
    subset_plugin.set_center((3, 2), update=True)
    subsets = cubeviz_helper.app.get_subsets()
    reg = subsets.get('Subset 1')[0]['region']
    assert_allclose(reg.center.x, 3)
    assert_allclose(reg.center.y, 2)
    assert_allclose(reg.width, 1.5)
    assert_allclose(reg.height, 3.3)
    assert_allclose(reg.angle.to_value(u.deg), 45)  # Might be stored in radians

    # Circular Subset
    subset_plugin.combination_mode = 'new'
    subset_plugin.import_region(CircularROI(xc=3, yc=4, radius=2.4))

    assert subset_plugin._obj.subset_selected == "Subset 2"
    assert subset_plugin._obj.subset_types == ["CircularROI"]
    assert subset_plugin._obj.is_centerable
    for key in ("orig", "value"):
        assert subset_plugin._obj._get_value_from_subset_definition(
            0, "X Center (pixels)", key) == 3
        assert subset_plugin._obj._get_value_from_subset_definition(
            0, "Y Center (pixels)", key) == 4
        assert subset_plugin._obj._get_value_from_subset_definition(
            0, "Radius (pixels)", key) == 2.4

    # Circular Annulus Subset
    subset_plugin.combination_mode = 'new'
    subset_plugin.import_region(CircularAnnulusROI(xc=5, yc=6, inner_radius=2, outer_radius=4))

    assert subset_plugin._obj.subset_selected == "Subset 3"
    assert subset_plugin._obj.subset_types == ["CircularAnnulusROI"]
    for key in ("orig", "value"):
        assert subset_plugin._obj._get_value_from_subset_definition(
            0, "X Center (pixels)", key) == 5
        assert subset_plugin._obj._get_value_from_subset_definition(
            0, "Y Center (pixels)", key) == 6
        assert subset_plugin._obj._get_value_from_subset_definition(
            0, "Inner Radius (pixels)", key) == 2
        assert subset_plugin._obj._get_value_from_subset_definition(
            0, "Outer Radius (pixels)", key) == 4


def test_region_from_subset_profile(cubeviz_helper, spectral_cube_wcs):
    data = Spectrum1D(flux=np.ones((128, 128, 256)) * u.nJy, wcs=spectral_cube_wcs)
    subset_plugin = cubeviz_helper.plugins['Subset Tools']._obj

    cubeviz_helper.load_data(data, data_label='Test 1D Flux')

    spectral_axis_unit = u.Unit(cubeviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    subset_plugin.import_region(SpectralRegion(5*spectral_axis_unit, 15.5*spectral_axis_unit))

    subsets = cubeviz_helper.app.get_subsets(spectral_only=True)
    reg = subsets.get('Subset 1')

    assert len(subsets) == 1
    assert isinstance(reg, SpectralRegion)
    assert_quantity_allclose(reg.lower, 5.0 * u.Hz)
    assert_quantity_allclose(reg.upper, 15.5 * u.Hz)

    assert subset_plugin.subset_selected == "Subset 1"
    assert subset_plugin.subset_types == ["Range"]
    assert subset_plugin.is_centerable
    assert_allclose(subset_plugin.get_center(), 10.25)
    for key in ("orig", "value"):
        assert subset_plugin._get_value_from_subset_definition(0, "Lower bound", key) == 5
        assert subset_plugin._get_value_from_subset_definition(0, "Upper bound", key) == 15.5

    # Mimic user changing something in Subset GUI.
    subset_plugin._set_value_in_subset_definition(0, "Lower bound", "value", 10)
    # "orig" is unchanged until user clicks Update button.
    assert subset_plugin._get_value_from_subset_definition(0, "Lower bound", "orig") == 5
    subset_plugin.vue_update_subset()
    for key in ("orig", "value"):
        assert subset_plugin._get_value_from_subset_definition(0, "Lower bound", key) == 10
        assert subset_plugin._get_value_from_subset_definition(0, "Upper bound", key) == 15.5

    subsets = cubeviz_helper.app.get_subsets(spectral_only=True)
    reg = subsets.get('Subset 1')

    assert_quantity_allclose(reg.lower, 10.0 * u.Hz)
    assert_quantity_allclose(reg.upper, 15.5 * u.Hz)

    # Move the Subset.
    subset_plugin.set_center(10, update=True)
    subsets = cubeviz_helper.app.get_subsets(spectral_only=True)
    reg = subsets.get('Subset 1')
    assert_quantity_allclose(reg.lower, 7.25 * u.Hz)
    assert_quantity_allclose(reg.upper, 12.75 * u.Hz)


def test_disjoint_spectral_subset(cubeviz_helper, spectral_cube_wcs):
    subset_plugin = cubeviz_helper.plugins['Subset Tools']._obj
    data = Spectrum1D(flux=np.ones((128, 128, 256)) * u.nJy, wcs=spectral_cube_wcs)
    cubeviz_helper.load_data(data, data_label="Test Flux")

    spectral_axis_unit = u.Unit(cubeviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    subset_plugin.import_region(SpectralRegion(5 * spectral_axis_unit, 15.5 * spectral_axis_unit))
    subset_plugin.combination_mode.selected = 'or'
    subset_plugin.import_region(SpectralRegion(30 * spectral_axis_unit, 35 * spectral_axis_unit))

    reg = cubeviz_helper.app.get_subsets('Subset 1')

    assert len(reg) == 2
    assert isinstance(reg, SpectralRegion)
    assert_quantity_allclose(reg[0].lower, 5.0*u.Hz)
    assert_quantity_allclose(reg[0].upper, 15.5*u.Hz)
    assert_quantity_allclose(reg[1].lower, 30.0*u.Hz)
    assert_quantity_allclose(reg[1].upper, 35.0*u.Hz)

    assert subset_plugin.subset_selected == "Subset 1"
    assert subset_plugin.subset_types == ["Range", "Range"]
    assert subset_plugin.glue_state_types == ["RangeSubsetState", "OrState"]

    # Make sure that certain things are not possible because we are
    # dealing with a composite spectral subset
    subset_plugin.set_center(99, update=True)   # This is no-op
    assert subset_plugin.get_center() is None

    for key in ("orig", "value"):
        assert subset_plugin._get_value_from_subset_definition(1, "Lower bound", key) == 30
        assert subset_plugin._get_value_from_subset_definition(1, "Upper bound", key) == 35
        assert subset_plugin._get_value_from_subset_definition(0, "Lower bound", key) == 5
        assert subset_plugin._get_value_from_subset_definition(0, "Upper bound", key) == 15.5

    # We will now update one of the bounds of the composite subset
    subset_plugin._set_value_in_subset_definition(1, "Lower bound", "value", 25)
    subset_plugin.vue_update_subset()
    assert subset_plugin._get_value_from_subset_definition(1, "Lower bound", "value") == 25
    assert subset_plugin._get_value_from_subset_definition(1, "Lower bound", "orig") == 25

    reg = cubeviz_helper.app.get_subsets('Subset 1')
    assert_quantity_allclose(reg[1].lower, 25.0*u.Hz)  # It is now the updated value


def test_composite_region_from_subset_3d(cubeviz_helper):
    cubeviz_helper.load_data(np.ones((128, 128, 10)), data_label='Test 3D Flux')

    cubeviz_helper.app.add_data_to_viewer('flux-viewer', 'Test 3D Flux')

    subset_plugin = cubeviz_helper.plugins['Subset Tools']
    subset_plugin.import_region(CircularROI(xc=25, yc=25, radius=5))

    reg = cubeviz_helper.app.get_subsets("Subset 1")
    circle1 = CirclePixelRegion(center=PixCoord(x=25, y=25), radius=5)
    assert reg[-1] == {'name': 'CircularROI', 'glue_state': 'RoiSubsetState', 'region': circle1,
                       'sky_region': None, 'subset_state': reg[-1]['subset_state']}

    subset_plugin.combination_mode.selected = 'andnot'
    subset_plugin.import_region(RectangularROI(25, 30, 25, 30))

    reg = cubeviz_helper.app.get_subsets("Subset 1")
    rectangle1 = RectanglePixelRegion(center=PixCoord(x=27.5, y=27.5),
                                      width=5, height=5, angle=0.0 * u.deg)
    assert reg[-1] == {'name': 'RectangularROI', 'glue_state': 'AndNotState', 'region': rectangle1,
                       'sky_region': None, 'subset_state': reg[-1]['subset_state']}

    subset_plugin.combination_mode.selected = 'or'
    subset_plugin.import_region(EllipticalROI(xc=30, yc=30, radius_x=3, radius_y=6))

    reg = cubeviz_helper.app.get_subsets("Subset 1")
    ellipse1 = EllipsePixelRegion(center=PixCoord(x=30, y=30),
                                  width=6, height=12, angle=0.0 * u.deg)
    assert reg[-1] == {'name': 'EllipticalROI', 'glue_state': 'OrState', 'region': ellipse1,
                       'sky_region': None, 'subset_state': reg[-1]['subset_state']}

    subset_plugin.combination_mode.selected = 'and'
    subset_plugin.import_region(RectangularROI(20, 25, 20, 25))

    reg = cubeviz_helper.app.get_subsets("Subset 1")
    rectangle2 = RectanglePixelRegion(center=PixCoord(x=22.5, y=22.5),
                                      width=5, height=5, angle=0.0 * u.deg)
    assert reg[-1] == {'name': 'RectangularROI', 'glue_state': 'AndState', 'region': rectangle2,
                       'sky_region': None, 'subset_state': reg[-1]['subset_state']}

    subset_plugin.combination_mode.selected = 'andnot'
    subset_plugin.import_region(CircularROI(xc=21, yc=24, radius=1))

    reg = cubeviz_helper.app.get_subsets("Subset 1")
    circle2 = CirclePixelRegion(center=PixCoord(x=21, y=24), radius=1)
    assert reg[-1] == {'name': 'CircularROI', 'glue_state': 'AndNotState', 'region': circle2,
                       'sky_region': None, 'subset_state': reg[-1]['subset_state']}

    subset_plugin = cubeviz_helper.app.get_tray_item_from_name('g-subset-tools')
    assert subset_plugin.subset_selected == "Subset 1"
    assert subset_plugin.subset_types == ['CircularROI', 'RectangularROI', 'EllipticalROI',
                                          'RectangularROI', 'CircularROI']
    assert subset_plugin.glue_state_types == ['AndState', 'AndNotState',
                                              'OrState', 'AndState', 'AndNotState']


def test_composite_region_with_consecutive_and_not_states(cubeviz_helper):
    cubeviz_helper.load_data(np.ones((128, 128, 10)), data_label='Test 3D Flux')
    subset_plugin = cubeviz_helper.plugins['Subset Tools']._obj

    cubeviz_helper.app.add_data_to_viewer('flux-viewer', 'Test 3D Flux')
    viewer = cubeviz_helper.app.get_viewer('flux-viewer')

    subset_plugin.import_region(CircularROI(xc=25, yc=25, radius=5))
    reg = cubeviz_helper.app.get_subsets("Subset 1")
    circle1 = CirclePixelRegion(center=PixCoord(x=25, y=25), radius=5)
    assert reg[-1] == {'name': 'CircularROI', 'glue_state': 'RoiSubsetState', 'region': circle1,
                       'sky_region': None, 'subset_state': reg[-1]['subset_state']}

    subset_plugin.combination_mode.selected = 'andnot'
    subset_plugin.import_region(RectangularROI(25, 30, 25, 30))
    reg = cubeviz_helper.app.get_subsets("Subset 1")
    rectangle1 = RectanglePixelRegion(center=PixCoord(x=27.5, y=27.5),
                                      width=5, height=5, angle=0.0 * u.deg)
    assert reg[-1] == {'name': 'RectangularROI', 'glue_state': 'AndNotState', 'region': rectangle1,
                       'sky_region': None, 'subset_state': reg[-1]['subset_state']}

    subset_plugin.combination_mode.selected = 'andnot'
    subset_plugin.import_region(EllipticalROI(xc=30, yc=30, radius_x=3, radius_y=6))
    reg = cubeviz_helper.app.get_subsets("Subset 1")
    ellipse1 = EllipsePixelRegion(center=PixCoord(x=30, y=30),
                                  width=6, height=12, angle=0.0 * u.deg)
    assert reg[-1] == {'name': 'EllipticalROI', 'glue_state': 'AndNotState', 'region': ellipse1,
                       'sky_region': None, 'subset_state': reg[-1]['subset_state']}

    regions_list = cubeviz_helper.app.get_subsets("Subset 1", object_only=True)
    assert len(regions_list) == 3
    assert regions_list[-1].width == 6

    regions_list = cubeviz_helper.app.get_subsets("Subset 1", spatial_only=True,
                                                  object_only=True)
    assert len(regions_list) == 3
    assert regions_list[-1].width == 6

    spatial_list = cubeviz_helper.app.get_subsets("Subset 1", spatial_only=True)
    assert len(spatial_list) == 3

    assert subset_plugin.subset_selected == "Subset 1"
    assert subset_plugin.subset_types == ['CircularROI', 'RectangularROI', 'EllipticalROI']
    assert subset_plugin.glue_state_types == ['AndState', 'AndNotState', 'AndNotState']

    # This should be prevented since radius must be positive
    subset_plugin._set_value_in_subset_definition(0, "Radius", "value", 0)
    subset_plugin.vue_update_subset()

    # This should also be prevented since a rectangle must have positive width
    # and length
    subset_plugin._set_value_in_subset_definition(1, "Xmin", "value", 0)
    subset_plugin._set_value_in_subset_definition(1, "Xmax", "value", 0)
    subset_plugin.vue_update_subset()

    # Make sure changes were not propagated
    reg = cubeviz_helper.app.get_subsets("Subset 1")
    assert reg[0]['subset_state'].roi.radius == 5
    assert reg[1]['subset_state'].roi.xmin == 25
    assert reg[1]['subset_state'].roi.xmax == 30

    for layer in viewer.state.layers:
        if isinstance(layer.layer, GroupedSubset):
            assert get_subset_type(layer.layer.subset_state) == 'spatial'


def test_composite_region_with_imviz(imviz_helper, image_2d_wcs):
    arr = NDData(np.ones((10, 10)), wcs=image_2d_wcs)
    subset_plugin = imviz_helper.plugins['Subset Tools']._obj
    data_label = 'image-data'
    imviz_helper.load_data(arr, data_label=data_label, show_in_viewer=True)
    subset_plugin.import_region(CircularROI(xc=5, yc=5, radius=2))
    reg = imviz_helper.app.get_subsets("Subset 1")
    circle1 = CirclePixelRegion(center=PixCoord(x=5, y=5), radius=2)
    assert reg[-1] == {'name': 'CircularROI', 'glue_state': 'RoiSubsetState', 'region': circle1,
                       'sky_region': None, 'subset_state': reg[-1]['subset_state']}

    subset_plugin.combination_mode.selected = 'andnot'
    subset_plugin.import_region(RectangularROI(xmin=2, xmax=4, ymin=2, ymax=4))
    reg = imviz_helper.app.get_subsets("Subset 1")
    rectangle1 = RectanglePixelRegion(center=PixCoord(x=3, y=3),
                                      width=2, height=2, angle=0.0 * u.deg)
    assert reg[-1] == {'name': 'RectangularROI', 'glue_state': 'AndNotState', 'region': rectangle1,
                       'sky_region': None, 'subset_state': reg[-1]['subset_state']}

    subset_plugin.combination_mode.selected = 'andnot'
    subset_plugin.import_region(EllipticalROI(xc=3, yc=3, radius_x=3, radius_y=6))
    reg = imviz_helper.app.get_subsets("Subset 1")
    ellipse1 = EllipsePixelRegion(center=PixCoord(x=3, y=3),
                                  width=6, height=12, angle=0.0 * u.deg)
    assert reg[-1] == {'name': 'EllipticalROI', 'glue_state': 'AndNotState', 'region': ellipse1,
                       'sky_region': None, 'subset_state': reg[-1]['subset_state']}

    subset_plugin.combination_mode.selected = 'or'
    subset_plugin.import_region(CircularAnnulusROI(xc=5, yc=5, inner_radius=2.5, outer_radius=5))
    reg = imviz_helper.app.get_subsets("Subset 1")
    ann1 = CircleAnnulusPixelRegion(center=PixCoord(x=5, y=5), inner_radius=2.5, outer_radius=5)
    assert reg[-1] == {'name': 'CircularAnnulusROI', 'glue_state': 'OrState', 'region': ann1,
                       'sky_region': None, 'subset_state': reg[-1]['subset_state']}

    assert subset_plugin.subset_selected == "Subset 1"
    assert subset_plugin.subset_types == ['CircularROI', 'RectangularROI', 'EllipticalROI',
                                          'CircularAnnulusROI']
    assert subset_plugin.glue_state_types == ['AndState', 'AndNotState', 'AndNotState', 'OrState']


def test_recenter_linked_by_wcs(imviz_helper):
    """Similar test case as TestAdvancedAperPhot.
    Images are only aligned if linked by WCS.
    """
    # Reference image
    imviz_helper.load_data(get_pkg_data_filename(
        'data/gauss100_fits_wcs.fits', package='jdaviz.configs.imviz.tests'))
    # Different pixel scale
    imviz_helper.load_data(get_pkg_data_filename(
        'data/gauss100_fits_wcs_block_reduced.fits', package='jdaviz.configs.imviz.tests'))

    # Link them by WCS
    imviz_helper.link_data(align_by='wcs')
    w = imviz_helper.app.data_collection[0].coords

    # This rectangle is over a real object in reference image but
    # only the last row in the second image if linked by pixel.
    imviz_helper.plugins['Subset Tools'].import_region(
        RectanglePixelRegion(center=PixCoord(x=229, y=152), width=17, height=7).to_sky(w))

    subset_plugin = imviz_helper.plugins['Subset Tools']
    subset_plugin.subset = "Subset 1"
    subset_plugin.recenter_dataset = "gauss100_fits_wcs_block_reduced[PRIMARY,1]"

    # Do it a few times to converge.
    for _ in range(5):
        subset_plugin.recenter()

    # If handled correctly, it won't change much.
    # But if not, it move down by 7 pix or so (229.05, 145.92) and fails the test.
    xy = imviz_helper.default_viewer._obj._get_real_xy(
        imviz_helper.app.data_collection[0], *subset_plugin.get_center())[:2]
    assert_allclose(xy, (229.067822, 152.371943))

    # Now create a new subset that has a source in the corner and test
    # recentering with multiselect.
    subset_plugin.combination_mode = 'new'
    subset_plugin.import_region(
        CirclePixelRegion(center=PixCoord(x=145, y=175), radius=17).to_sky(w))
    subset_plugin._obj.multiselect = True
    subset_plugin._obj.subset_selected = ["Subset 1", "Subset 2"]

    # Do it a few times to converge.
    for _ in range(5):
        subset_plugin._obj.recenter()

    xy = imviz_helper.default_viewer._obj._get_real_xy(
        imviz_helper.app.data_collection[0], *subset_plugin.get_center("Subset 2"))[:2]
    assert_allclose(xy, (145.593022, 172.515541))

    with pytest.raises(ValueError, match="Please include subset_name in"):
        subset_plugin.get_center()
    with pytest.raises(ValueError, match="Please include subset_name in"):
        subset_plugin.set_center((150, 200))


def test_with_invalid_subset_name(cubeviz_helper):
    subset_name = "Test"
    with pytest.raises(ValueError, match=f'{subset_name} not in '):
        cubeviz_helper.app.get_subsets(subset_name=subset_name)


def test_composite_region_from_subset_2d(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)
    viewer = specviz_helper.app.get_viewer(specviz_helper._default_spectrum_viewer_reference_name)
    subset_plugin = specviz_helper.plugins['Subset Tools']._obj

    subset1 = SpectralRegion(6000 * spectrum1d.spectral_axis.unit,
                             7000 * spectrum1d.spectral_axis.unit)
    subset_plugin.import_region(subset1)
    reg = specviz_helper.app.get_subsets("Subset 1", simplify_spectral=False)

    assert reg[-1]['region'].lower == subset1.lower and reg[-1]['region'].upper == subset1.upper
    assert reg[-1]['glue_state'] == 'RangeSubsetState'

    subset_plugin.combination_mode.selected = 'andnot'
    subset1 = SpectralRegion(6500 * spectrum1d.spectral_axis.unit,
                             6800 * spectrum1d.spectral_axis.unit)
    subset_plugin.import_region(subset1)
    reg = specviz_helper.app.get_subsets("Subset 1", simplify_spectral=False)

    assert reg[-1]['region'].lower == subset1.lower and reg[-1]['region'].upper == subset1.upper
    assert reg[-1]['glue_state'] == 'AndNotState'

    subset_plugin.combination_mode.selected = 'or'
    subset1 = SpectralRegion(7200 * spectrum1d.spectral_axis.unit,
                             7800 * spectrum1d.spectral_axis.unit)
    subset_plugin.import_region(subset1)

    reg = specviz_helper.app.get_subsets("Subset 1", simplify_spectral=False)

    assert reg[-1]['region'].lower == subset1.lower and reg[-1]['region'].upper == subset1.upper
    assert reg[-1]['glue_state'] == 'OrState'

    subset_plugin.combination_mode.selected = 'and'
    subset1 = SpectralRegion(6800 * spectrum1d.spectral_axis.unit,
                             7500 * spectrum1d.spectral_axis.unit)
    subset_plugin.import_region(subset1)

    reg = specviz_helper.app.get_subsets("Subset 1", simplify_spectral=False)

    assert reg[-1]['region'].lower == subset1.lower and reg[-1]['region'].upper == subset1.upper
    assert reg[-1]['glue_state'] == 'AndState'

    assert subset_plugin.subset_selected == "Subset 1"
    assert subset_plugin.subset_types == ['Range', 'Range', 'Range', 'Range']
    assert subset_plugin.glue_state_types == ['AndState', 'AndNotState', 'OrState', 'AndState']

    subset_plugin.vue_simplify_subset()
    assert subset_plugin.glue_state_types == ["RangeSubsetState", "OrState"]

    for layer in viewer.state.layers:
        if isinstance(layer.layer, GroupedSubset):
            assert get_subset_type(layer.layer.subset_state) == 'spectral'


def test_edit_composite_spectral_subset(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)
    subset_plugin = specviz_helper.plugins['Subset Tools']._obj

    unit = spectrum1d.spectral_axis.unit
    subset = [SpectralRegion(6200 * unit, 6800 * unit),
              SpectralRegion(7200 * unit, 7600 * unit),
              SpectralRegion(6200 * unit, 7600 * unit)]
    mode = ['new', 'or', 'xor']
    subset_plugin.import_region(region=subset, combination_mode=mode)

    reg = specviz_helper.app.get_subsets("Subset 1")
    assert reg.lower.value == 6800 and reg.upper.value == 7200

    # We will now update one of the bounds of the composite subset
    subset_plugin._set_value_in_subset_definition(0, "Lower bound", "value", 6000)
    subset_plugin.vue_update_subset()

    # Since we updated one of the Range objects and it's lower bound
    # is now lower than the XOR region bound, the region from 6000 to
    # 6200 should now be visible in the viewer.
    reg = specviz_helper.app.get_subsets("Subset 1")
    assert reg[0].lower.value == 6000 and reg[0].upper.value == 6200
    assert reg[1].lower.value == 6800 and reg[1].upper.value == 7200

    # This makes it so that only spectral regions within this bound
    # are visible, so the API should reflect that.
    subset_plugin.combination_mode.selected = 'and'
    subset_plugin.import_region(SpectralRegion(6600 * unit, 7400 * unit))

    reg = specviz_helper.app.get_subsets("Subset 1")
    assert reg.lower.value == 6800 and reg[0].upper.value == 7200

    # This should be prevented by the _check_inputs method
    subset_plugin._set_value_in_subset_definition(0, "Lower bound", "value", 8000)
    subset_plugin.vue_update_subset()
    reg2 = specviz_helper.app.get_subsets("Subset 1")
    assert reg.lower.value == reg2.lower.value
    assert reg.upper.value == reg2.upper.value

    assert subset_plugin.can_simplify

    subset_plugin.combination_mode.selected = 'and'
    subset_plugin.import_region(SpectralRegion(7800 * unit, 8000 * unit))
    with pytest.raises(ValueError, match="AND mode should overlap with existing subset"):
        specviz_helper.app.get_subsets("Subset 1")


def test_composite_spectral_with_xor(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)

    subset_plugin = specviz_helper.plugins['Subset Tools']

    unit = spectrum1d.spectral_axis.unit
    subset = [SpectralRegion(6200 * unit, 6800 * unit),
              SpectralRegion(7200 * unit, 7600 * unit),
              SpectralRegion(6100 * unit, 7600 * unit)]
    mode = ['new', 'or', 'xor']
    subset_plugin.import_region(region=subset, combination_mode=mode)

    reg = specviz_helper.app.get_subsets("Subset 1")

    assert reg[0].lower.value == 6100 and reg[0].upper.value == 6200
    assert reg[1].lower.value == 6800 and reg[1].upper.value == 7200

    subset_plugin.combination_mode = 'new'
    subset = [SpectralRegion(7000 * unit, 7200 * unit),
              SpectralRegion(7100 * unit, 7300 * unit),
              SpectralRegion(6900 * unit, 7105 * unit)]
    mode = ['new', 'xor', 'or']
    subset_plugin.import_region(region=subset, combination_mode=mode)

    reg = specviz_helper.app.get_subsets("Subset 2")
    assert reg[0].lower.value == 6900 and reg[0].upper.value == 7105
    assert reg[1].lower.value == 7200 and reg[1].upper.value == 7300

    subset_plugin.combination_mode = 'new'
    subset = [SpectralRegion(6000 * unit, 6500 * unit),
              SpectralRegion(6100 * unit, 6200 * unit)]
    mode = ['new', 'xor']
    subset_plugin.import_region(region=subset, combination_mode=mode)

    reg = specviz_helper.app.get_subsets("Subset 3")
    assert reg[0].lower.value == 6000 and reg[0].upper.value == 6100
    assert reg[1].lower.value == 6200 and reg[1].upper.value == 6500

    subset_plugin.combination_mode = 'new'
    subset = [SpectralRegion(6100 * unit, 6200 * unit),
              SpectralRegion(6000 * unit, 6500 * unit)]
    mode = ['new', 'xor']
    subset_plugin.import_region(region=subset, combination_mode=mode)

    reg = specviz_helper.app.get_subsets("Subset 4")
    assert reg[0].lower.value == 6000 and reg[0].upper.value == 6100
    assert reg[1].lower.value == 6200 and reg[1].upper.value == 6500

    subset_plugin.combination_mode = 'new'
    subset = [SpectralRegion(7500 * unit, 7600 * unit),
              SpectralRegion(6000 * unit, 6010 * unit)]
    mode = ['new', 'xor']
    subset_plugin.import_region(region=subset, combination_mode=mode)

    reg = specviz_helper.app.get_subsets("Subset 5")
    assert reg[0].lower.value == 6000 and reg[0].upper.value == 6010
    assert reg[1].lower.value == 7500 and reg[1].upper.value == 7600


def test_composite_spectral_with_xor_complicated(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)

    subset_plugin = specviz_helper.plugins['Subset Tools']

    unit = spectrum1d.spectral_axis.unit
    subset = [SpectralRegion(6100 * unit, 6700 * unit),
              SpectralRegion(6200 * unit, 6300 * unit),
              SpectralRegion(6050 * unit, 6800 * unit),
              SpectralRegion(7000 * unit, 7200 * unit),
              SpectralRegion(6010 * unit, 6020 * unit),
              SpectralRegion(6090 * unit, 6850 * unit)
              ]
    mode = ['new', 'andnot', 'xor', 'or', 'or', 'xor']
    subset_plugin.import_region(region=subset, combination_mode=mode)

    # (6100, 6200), (6300, 6700)
    # (6050, 6100), (6200, 6300), (6700, 6800)
    # (6050, 6100), (6200, 6300), (6700, 6800), (7000, 7200)
    # (6010, 6020), (6050, 6100), (6200, 6300), (6700, 6800), (7000, 7200)
    # (6010, 6020), (6050, 6090), (6100, 6200), (6300, 6700), (6800, 6850), (7000, 7200)

    reg = specviz_helper.app.get_subsets("Subset 1")
    assert reg[0].lower.value == 6010 and reg[0].upper.value == 6020
    assert reg[1].lower.value == 6050 and reg[1].upper.value == 6090
    assert reg[2].lower.value == 6100 and reg[2].upper.value == 6200
    assert reg[3].lower.value == 6300 and reg[3].upper.value == 6700
    assert reg[4].lower.value == 6800 and reg[4].upper.value == 6850
    assert reg[5].lower.value == 7000 and reg[5].upper.value == 7200


def test_overlapping_spectral_regions(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)
    subset_plugin = specviz_helper.plugins['Subset Tools']._obj
    unit = spectrum1d.spectral_axis.unit
    subset = [SpectralRegion(6400 * unit, 7400 * unit),
              SpectralRegion(6600 * unit, 7200 * unit),
              SpectralRegion(6600 * unit, 7300 * unit)]
    mode = ['new', 'andnot', 'or']
    subset_plugin.import_region(region=subset, combination_mode=mode)

    assert subset_plugin.can_simplify
    subset_plugin.vue_simplify_subset()

    reg = specviz_helper.app.get_subsets("Subset 1")
    assert reg.lower.value == 6400 and reg.upper.value == 7400


def test_only_overlapping_spectral_regions(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)
    subset_plugin = specviz_helper.plugins['Subset Tools']._obj
    unit = spectrum1d.spectral_axis.unit
    subset_plugin.import_region(SpectralRegion(6400 * unit, 6600 * unit))

    assert specviz_helper.app.is_there_overlap_spectral_subset("Subset 1") is False

    subset_plugin.combination_mode.selected = 'or'
    subset_plugin.import_region(SpectralRegion(7000 * unit, 7400 * unit))
    subset_plugin.import_region(SpectralRegion(6600 * unit, 7300 * unit))
    subset_plugin.import_region(SpectralRegion(7600 * unit, 7800 * unit))

    assert subset_plugin.can_simplify
    subset_plugin.vue_simplify_subset()

    reg = specviz_helper.app.get_subsets("Subset 1")
    assert reg[0].lower.value == 6400 and reg[0].upper.value == 7400
    assert reg[1].lower.value == 7600 and reg[1].upper.value == 7800


def test_overlapping_in_specviz2d(specviz2d_helper, mos_spectrum2d):
    specviz2d_helper.load_data(spectrum_2d=mos_spectrum2d)

    subset_plugin = specviz2d_helper.plugins['Subset Tools']._obj
    unit = mos_spectrum2d.spectral_axis.unit

    subset_plugin.import_region(SpectralRegion(6400 * unit, 7400 * unit))
    subset_plugin.combination_mode.selected = 'andnot'
    subset_plugin.import_region(SpectralRegion(6600 * unit, 7200 * unit))
    subset_plugin.combination_mode.selected = 'or'
    subset_plugin.import_region(SpectralRegion(6600 * unit, 7300 * unit))

    assert subset_plugin.can_simplify
    subset_plugin.vue_simplify_subset()

    reg = specviz2d_helper.app.get_subsets("Subset 1")
    assert reg.lower.value == 6400 and reg.upper.value == 7400


def test_only_overlapping_in_specviz2d(specviz2d_helper, mos_spectrum2d):
    specviz2d_helper.load_data(spectrum_2d=mos_spectrum2d)

    subset_plugin = specviz2d_helper.plugins['Subset Tools']._obj
    unit = mos_spectrum2d.spectral_axis.unit

    subset_plugin.import_region(SpectralRegion(6400 * unit, 6600 * unit))
    subset_plugin.combination_mode.selected = 'or'
    subset_plugin.import_region(SpectralRegion(7000 * unit, 7400 * unit))
    subset_plugin.import_region(SpectralRegion(6600 * unit, 7300 * unit))
    subset_plugin.import_region(SpectralRegion(7600 * unit, 7800 * unit))

    assert subset_plugin.can_simplify
    subset_plugin.vue_simplify_subset()

    reg = specviz2d_helper.app.get_subsets("Subset 1")
    assert reg[0].lower.value == 6400 and reg[0].upper.value == 7400
    assert reg[1].lower.value == 7600 and reg[1].upper.value == 7800


def test_draw2d_linking_specviz2d(specviz2d_helper):
    # custom test data to predict values for different viewers
    header = {
              'WCSAXES': 2,
              'CRPIX1': 0.0, 'CRPIX2': 8.5,
              'CDELT1': 1E-06, 'CDELT2': 7.5E-05,
              'CUNIT1': 'm', 'CUNIT2': 'deg',
              'CTYPE1': 'WAVE', 'CTYPE2': 'OFFSET',
              'CRVAL1': 0.0, 'CRVAL2': 5.0,
              'RADESYS': 'ICRS', 'SPECSYS': 'BARYCENT'}
    wcs = WCS(header)

    x_values = np.linspace(0, 10, 128)
    y_values = np.linspace(0, 5, 256)

    # Create a continuous 2D
    data = np.sin(x_values[:, np.newaxis]) * np.cos(y_values) * u.one
    spectrum_data = Spectrum1D(data, wcs=wcs, meta=header)

    specviz2d_helper.load_data(spectrum_2d=spectrum_data)
    viewer_1d = specviz2d_helper.app.get_viewer(
        specviz2d_helper._default_spectrum_viewer_reference_name)
    viewer_2d = specviz2d_helper.app.get_viewer(
        specviz2d_helper._default_spectrum_2d_viewer_reference_name)

    # create subset in 2d viewer,  want data in 1d viewer
    viewer_2d.apply_roi(XRangeROI(60, 80))
    subset_drawn_2d = viewer_1d.native_marks[-1]

    # get x and y components to compute subset mask
    y1 = subset_drawn_2d.y
    x1 = subset_drawn_2d.x

    subset_highlighted_region1 = x1[np.isfinite(y1)]
    min_value_subset1 = np.min(subset_highlighted_region1)
    max_value_subset1 = np.max(subset_highlighted_region1)

    tolerance1 = 1e-6
    expected_min1 = 6.e-5
    expected_max1 = 8.e-05

    assert np.allclose(min_value_subset1, expected_min1, atol=tolerance1)
    assert np.allclose(max_value_subset1, expected_max1, atol=tolerance1)

    # now create a subset in the spectrum-viewer, and determine if
    # subset is linked correctly in spectrum2d-viewer
    spec_reg = SpectralRegion(0.0001 * u.m, .0002 * u.m)
    st = specviz2d_helper.plugins['Subset Tools']
    st.import_region(spec_reg)

    mask = viewer_2d._get_layer('Subset 1')._get_image()
    x_coords = np.nonzero(mask)[1]
    min_value_subset2 = x_coords.min()
    max_value_subset2 = x_coords.max()

    tolerance2 = 1
    expected_min2 = 100
    expected_max2 = 199

    assert np.allclose(min_value_subset2, expected_min2, atol=tolerance2)
    assert np.allclose(max_value_subset2, expected_max2, atol=tolerance2)


def test_multi_mask_subset(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)
    subset_plugin = specviz_helper.plugins['Subset Tools']._obj
    unit = spectrum1d.spectral_axis.unit
    subset_plugin.import_region(SpectralRegion(6200 * unit, 6800 * unit))

    subset_plugin.can_freeze = True
    subset_plugin.vue_freeze_subset()

    reg = specviz_helper.app.get_subsets()
    assert reg["Subset 1"][0]["region"] == 3
    assert isinstance(reg["Subset 1"][0]["subset_state"], MultiMaskSubsetState)

    subset_plugin.combination_mode.selected = 'or'
    subset_plugin.import_region(SpectralRegion(7200 * unit, 7600 * unit))

    # Simplify subset ignores Mask subsets
    reg = specviz_helper.app.get_subsets()
    assert (reg["Subset 1"].lower.value == 7200
            and reg["Subset 1"].upper.value == 7600)

    # If we set simplify to False, we see all subregions
    reg = specviz_helper.app.get_subsets(simplify_spectral=False)
    assert (reg["Subset 1"][1]["region"].lower.value == 7200
            and reg["Subset 1"][1]["region"].upper.value == 7600)
    assert reg["Subset 1"][0]["region"] == 3
    assert subset_plugin.can_simplify is False

    # If we freeze again, all subregions become a Mask subset object
    subset_plugin.vue_freeze_subset()
    reg = specviz_helper.app.get_subsets()
    assert reg["Subset 1"][0]["region"] == 5

    # When freezing an AndNot state, the number of mask values should decrease
    subset_plugin.combination_mode.selected = 'andnot'
    subset_plugin.import_region(SpectralRegion(6600 * unit, 7200 * unit))

    reg = specviz_helper.app.get_subsets(simplify_spectral=False)
    assert (reg["Subset 1"][1]["region"].lower.value == 6600
            and reg["Subset 1"][1]["region"].upper.value == 7200)
    assert reg["Subset 1"][0]["region"] == 5

    subset_plugin.vue_freeze_subset()
    reg = specviz_helper.app.get_subsets()
    assert reg["Subset 1"][0]["region"] == 4


def test_delete_subsets_toolbar_selection(cubeviz_helper, spectral_cube_wcs):
    """
    Test that the toolbar selections get reset when the subset being actively edited gets deleted.
    """
    data = Spectrum1D(flux=np.ones((128, 128, 256)) * u.nJy, wcs=spectral_cube_wcs)
    cubeviz_helper.load_data(data, data_label="Test Flux")
    dc = cubeviz_helper.app.data_collection

    spectrum_viewer = cubeviz_helper.app.get_viewer("spectrum-viewer")

    subset_plugin = cubeviz_helper.plugins['Subset Tools']
    unit = u.Unit(cubeviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    subset_plugin.import_region(SpectralRegion(6200 * unit, 6800 * unit))

    dc.remove_subset_group(dc.subset_groups[0])

    assert spectrum_viewer.toolbar.active_tool_id == "jdaviz:selectslice"

    flux_viewer = cubeviz_helper.app.get_viewer("flux-viewer")

    subset_plugin.combination_mode = 'new'
    subset_plugin.import_region(RectangularROI(1, 3.5, -0.2, 3.3))

    dc.remove_subset_group(dc.subset_groups[0])

    assert flux_viewer.toolbar.active_tool is None


def test_delete_subsets_app_api(cubeviz_helper, spectral_cube_wcs):
    """Test app.delete_subsets."""

    data = Spectrum1D(flux=np.ones((128, 128, 256)) * u.nJy, wcs=spectral_cube_wcs)
    cubeviz_helper.load_data(data, data_label="Test Flux")
    dc = cubeviz_helper.app.data_collection
    subset_plugin = cubeviz_helper.plugins['Subset Tools']
    unit = u.Unit(cubeviz_helper.plugins['Unit Conversion'].spectral_unit.selected)

    # spectral and spatial region to be loaded as subsets
    spectral = SpectralRegion(6200 * unit, 6800 * unit)
    spatial = CirclePixelRegion(center=PixCoord(x=25, y=25), radius=5)

    # add one spatial and one spectral subset
    subset_plugin.import_region([spectral, spatial], combination_mode='new')
    assert len(dc.subset_groups) == 2  # initially 2 subsets
    cubeviz_helper.app.delete_subsets()  # no args, will delete all subsets
    assert len(dc.subset_groups) == 0  # after delete all

    # re-add both subsets, test deletion while specifying single label
    subset_plugin.import_region([spectral, spatial], combination_mode='new')
    assert len(dc.subset_groups) == 2
    cubeviz_helper.app.delete_subsets('Subset 3')
    assert len(dc.subset_groups) == 1
    assert dc.subset_groups[0].label == 'Subset 4'  # Subset 4 should remain

    # same as previous scenario, but single label is in list
    assert len(dc.subset_groups) == 1
    cubeviz_helper.app.delete_subsets(['Subset 4'])
    assert len(dc.subset_groups) == 0  # now no subsets remain

    # re-add both subsets, test deletion while specifying list of labels
    subset_plugin.import_region([spectral, spatial], combination_mode='new')
    assert len(dc.subset_groups) == 2
    cubeviz_helper.app.delete_subsets(['Subset 5', 'Subset 6'])
    assert len(dc.subset_groups) == 0  # both should have been deleted

    # test that the proper error is raised for non-existent subsets
    match_str = r'not in data collection, can not delete\.'
    with pytest.raises(ValueError, match='Subset 5 ' + match_str):
        cubeviz_helper.app.delete_subsets(['Subset 5'])

    with pytest.raises(ValueError, match='Subset 5, Subset 6 ' + match_str):
        cubeviz_helper.app.delete_subsets(['Subset 5', 'Subset 6'])


class TestRegionsFromSubsets:
    """Tests for obtaining Sky Regions from subsets."""

    def test_get_regions_from_subsets_cubeviz(self, cubeviz_helper, spectral_cube_wcs):
        """ Basic tests for retrieving Sky Regions from spatial subsets in Cubeviz.
        """
        data = Spectrum1D(flux=np.ones((128, 128, 256)) * u.nJy, wcs=spectral_cube_wcs)
        cubeviz_helper.load_data(data)

        # basic test, a single circular region
        subset_plugin = cubeviz_helper.plugins['Subset Tools']
        subset_plugin.import_region(CircularROI(25, 25, 10))
        subsets = cubeviz_helper.app.get_subsets(include_sky_region=True)
        sky_region = subsets['Subset 1'][0]['sky_region']
        assert isinstance(sky_region, CircleSkyRegion)
        assert_allclose(sky_region.center.ra.deg, 24.40786313)
        assert_allclose(sky_region.center.dec.deg, 22.45185308)
        assert_allclose(sky_region.radius.arcsec, 28001.08106569353)

        # and that it is None when not specified
        subsets = cubeviz_helper.app.get_subsets()
        assert subsets['Subset 1'][0]['sky_region'] is None

        # now test a composite subset, each component should have a sky region
        subset_plugin.combination_mode.selected = 'and'
        subset_plugin.import_region(CircularROI(30, 30, 10))

        subsets = cubeviz_helper.app.get_subsets(include_sky_region=True)
        assert len(subsets['Subset 1']) == 2
        sky_region_0 = subsets['Subset 1'][0]['sky_region']
        sky_region_1 = subsets['Subset 1'][1]['sky_region']
        assert_allclose(sky_region_0.center.ra.deg, 24.40786313)
        assert_allclose(sky_region_0.center.dec.deg, 22.45185308)
        assert_allclose(sky_region_0.radius.arcsec, 28001.08106569353)
        assert_allclose(sky_region_1.center.ra.deg, 28.41569583)
        assert_allclose(sky_region_1.center.dec.deg, 25.44814949)
        assert_allclose(sky_region_1.radius.arcsec, 25816.498273)

        # and that they are both None when not specified
        subsets = cubeviz_helper.app.get_subsets()
        assert subsets['Subset 1'][0]['sky_region'] is None
        assert subsets['Subset 1'][1]['sky_region'] is None

    def test_get_regions_from_subsets_imviz(self, imviz_helper, spectral_cube_wcs):
        """ Basic tests for retrieving Sky Regions from subsets in Imviz.
        """

        # using cube WCS instead of 2d imaging wcs for consistancy with
        # cubeviz test. accessing just the spatial part of this.
        wcs = spectral_cube_wcs.celestial

        data = NDData(np.ones((128, 128)) * u.nJy, wcs=wcs)
        imviz_helper.load_data(data)

        # basic test, a single circular region
        subset_plugin = imviz_helper.plugins['Subset Tools']
        subset_plugin.import_region(CircularROI(25, 25, 10))
        subsets = imviz_helper.app.get_subsets(include_sky_region=True)
        sky_region = subsets['Subset 1'][0]['sky_region']
        assert isinstance(sky_region, CircleSkyRegion)
        assert_allclose(sky_region.center.ra.deg, 24.40786313)
        assert_allclose(sky_region.center.dec.deg, 22.45185308)
        assert_allclose(sky_region.radius.arcsec, 28001.08106569353)

        # now test a composite subset, each component should have a sky region
        subset_plugin.combination_mode.selected = 'and'
        subset_plugin.import_region(CircularROI(30, 30, 10))

        subsets = imviz_helper.app.get_subsets(include_sky_region=True)
        assert len(subsets['Subset 1']) == 2
        sky_region_0 = subsets['Subset 1'][0]['sky_region']
        sky_region_1 = subsets['Subset 1'][1]['sky_region']
        assert_allclose(sky_region_0.center.ra.deg, 24.40786313)
        assert_allclose(sky_region_0.center.dec.deg, 22.45185308)
        assert_allclose(sky_region_0.radius.arcsec, 28001.08106569353)
        assert_allclose(sky_region_1.center.ra.deg, 28.41569583)
        assert_allclose(sky_region_1.center.dec.deg, 25.44814949)
        assert_allclose(sky_region_1.radius.arcsec, 25816.498273)

    def test_no_wcs_sky_regions(self, imviz_helper):
        """ Make sure that if sky regions are requested and there is no WCS,
            that it returns None with no error.
        """

        data = NDData(np.ones((40, 40)) * u.nJy)
        imviz_helper.load_data(data)

        subset_plugin = imviz_helper.plugins['Subset Tools']
        subset_plugin.import_region(CircularROI(25, 25, 10))
        subsets = imviz_helper.app.get_subsets(include_sky_region=True)
        assert subsets['Subset 1'][0]['sky_region'] is None

    def test_subset_renaming(self, specviz_helper, spectrum1d):
        specviz_helper.load_data(spectrum1d, 'myfile')

        subset_plugin = specviz_helper.plugins['Subset Tools']
        subset_plugin.import_region(SpectralRegion(6200 * spectrum1d.spectral_axis.unit,
                                                   7200 * spectrum1d.spectral_axis.unit))
        get_data_no_sub = specviz_helper.get_data('myfile')
        get_data_1 = specviz_helper.get_data('myfile', spectral_subset='Subset 1')

        get_data_1_mask = np.where(~get_data_1.mask)

        # Retrieving data with no subset means mask is None
        assert get_data_no_sub.mask is None
        assert len(get_data_1_mask[0]) > 0

        # rename subset to 'diffname'
        subset_group = specviz_helper.app.data_collection.subset_groups
        subset_group[0].label = 'diffname'
        get_data_2 = specviz_helper.get_data('myfile', spectral_subset='diffname')

        assert_quantity_allclose(get_data_1.flux, get_data_2.flux)
        assert_quantity_allclose(get_data_1.spectral_axis, get_data_2.spectral_axis)
        get_data_2_mask = np.where(~get_data_2.mask)
        assert (get_data_1_mask[0] == get_data_2_mask[0]).all()

        # when we officially implement subset renaming in jdaviz, we will want
        # to be aware of the case of naming subsets after existing data
        # and handle it appropriately
        subset_group = specviz_helper.app.data_collection.subset_groups
        subset_group[0].label = 'myfile'
        get_data_3 = specviz_helper.get_data('myfile', spectral_subset='myfile')

        get_data_3_mask = np.where(~get_data_3.mask)
        assert (get_data_1_mask[0] == get_data_3_mask[0]).all()
